import gym
import numpy as np
import scipy.stats as stats
import torch
import torch.nn as nn
import torch.optim as optim

import pytorchrl as prl
from pytorchrl.agent.algorithms.base import Algorithm
from pytorchrl.agent.algorithms.utils import get_gradients, set_gradients


class MPC_CEM(Algorithm):
    """
    Model-Based MPC Cross-Entropy Method (CEM) class.
    Trains a model of the environment and uses CEM to select actions.

    Parameters
    ----------
    lr: float
        Dynamics model learning rate.
    envs : VecEnv
        Vector of environments instance.
    actor : Actor
        actor class instance.
    device : torch.device
        CPU or specific GPU where class computations will take place.
    mb_epochs : int
        Training epochs for the dynamics model.
    start_steps: int
        Number of steps collected with initial random policy.
    update_every : int
         Amount of data collected in between dynamics model updates.
    action_noise :
        Exploration noise.
    mini_batch_size : int
        Size of actor update batches.
    ub : float
        Actions upper bound.
    lb : float
        Actions lower bound.
    k_best : int
        Number of best action proposals per iteration.
    epsilon : float
        Threshold to stop the training iteration earlier if the action variance is very low.
    update_alpha :
        Action distribution mean soft update parameter.
    iter_update_steps :
        Number of optimizing action sampling iterations.
    max_grad_norm : float
        Gradient clipping parameter.
    test_every : int
        Regularity of test evaluations.
    num_test_episodes : int
        Number of episodes to complete in each test phase.
    """

    def __init__(self,
                 lr,
                 envs,
                 actor,
                 device,
                 start_steps,
                 update_every,
                 mb_epochs,
                 action_noise,
                 mini_batch_size,
                 ub=1,
                 lb=-1,
                 k_best=5,
                 epsilon=0.001,
                 update_alpha=0.0,
                 max_grad_norm=0.5,
                 iter_update_steps=3,
                 test_every=10,
                 num_test_episodes=3):

        # ---- General algo attributes ----------------------------------------

        # Number of steps collected with initial random policy
        self._start_steps = int(start_steps)

        # Times data in the buffer is re-used before data collection proceeds
        self._num_epochs = int(mb_epochs)

        # Tracks the number of times data is reused
        self.mb_train_epochs = 0

        # Number of data samples collected between network update stages
        self._update_every = int(update_every)

        # Number mini batches per epoch
        self._num_mini_batch = int(1)  # Depends on how much data is available

        # Size of update mini batches
        self._mini_batch_size = int(mini_batch_size)

        # Number of network updates between test evaluations
        self._test_every = int(test_every)

        # Number of episodes to complete when testing
        self._num_test_episodes = num_test_episodes

        # ---- CEM-specific attributes ----------------------------------------

        # Number of episodes to complete when testing
        self.iter = 0
        self.envs = envs
        self.actor = actor
        self.device = device
        self.max_grad_norm = max_grad_norm
        self.reuse_data = False
        self.action_noise = action_noise

        assert isinstance(self.actor.dynamics_model.action_space, gym.spaces.Box),\
            "CEM requires a continuous action space!"

        self.lb = lb
        self.ub = ub
        self.k_best = k_best
        self.epsilon = epsilon
        self.update_alpha = update_alpha
        self.iter_update_steps = iter_update_steps

        # ----- Optimizers ----------------------------------------------------

        self.dynamics_optimizer = optim.Adam(self.actor.dynamics_model.parameters(), lr=lr)
        self.loss_func = torch.nn.MSELoss()

    @classmethod
    def create_factory(cls,
                       lr,
                       start_steps,
                       update_every,
                       mb_epochs,
                       action_noise,
                       mini_batch_size,
                       ub=1,
                       lb=-1,
                       k_best=5,
                       epsilon=0.001,
                       update_alpha=0.0,
                       max_grad_norm=0.5,
                       iter_update_steps=3,
                       test_every=10,
                       num_test_episodes=3):
        """
        Returns a function to create a new Model-Based MPC instance.

        lr: float
            Dynamics model learning rate.
        mb_epochs : int
            Training epochs for the dynamics model.
        start_steps: int
            Number of steps collected with initial random policy.
        update_every : int
             Amount of data collected in between dynamics model updates.
        action_noise :
            Exploration noise.
        mini_batch_size : int
            Size of actor update batches.
        ub : float
            Actions upper bound.
        lb : float
            Actions lower bound.
        k_best : int
            Number of best action proposals per iteration.
        epsilon : float
            Threshold to stop the training iteration earlier if the action variance is very low.
        update_alpha :
            Action distribution mean soft update parameter.
        iter_update_steps :
            Number of optimizing action sampling iterations.
        max_grad_norm : float
            Gradient clipping parameter.
        test_every : int
            Regularity of test evaluations.
        num_test_episodes : int
            Number of episodes to complete in each test phase.

        Returns
        -------
        create_algo_instance : func
            Function that creates a new MPC_CEM class instance.
        algo_name : str
            Name of the algorithm.
        """

        def create_algo_instance(device, actor, envs):
            return cls(lr=lr,
                       envs=envs,
                       actor=actor,
                       device=device,
                       mb_epochs=update_every,
                       start_steps=start_steps,
                       update_every=update_every,
                       action_noise=action_noise,
                       mini_batch_size=mini_batch_size,
                       test_every=test_every,

                       k_best=k_best,
                       update_alpha=update_alpha,
                       iter_update_steps=iter_update_steps,

                       lb=lb,
                       ub=ub,
                       epsilon=epsilon,
                       max_grad_norm=max_grad_norm,
                       )

        return create_algo_instance, prl.MPC_RS

    @property
    def gamma(self):
        """Returns discount factor gamma."""
        return None

    @property
    def start_steps(self):
        """Returns the number of steps to collect with initial random policy."""
        return self._start_steps

    @property
    def num_epochs(self):
        """
        Returns the number of times the whole buffer is re-used before data
        collection proceeds.
        """
        return self._num_epochs

    @property
    def update_every(self):
        """
        Returns the number of data samples collected between
        network update stages.
        """
        return self._update_every

    @property
    def num_mini_batch(self):
        """
        Returns the number of times the whole buffer is re-used before data
        collection proceeds.
        """
        return self._num_mini_batch

    @property
    def mini_batch_size(self):
        """
        Returns the number of mini batches per epoch.
        """
        return self._mini_batch_size

    @property
    def test_every(self):
        """Number of network updates between test evaluations."""
        return self._test_every

    @property
    def num_test_episodes(self):
        """
        Returns the number of episodes to complete when testing.
        """
        return self._num_test_episodes

    def select_k_best(self, rewards, action_hist):
        """Selects k action trajectories that led to the highest reward.

        Parameters
        ----------
        rewards: np.array
            Rewards per rollout
        action_history: np.array
            Action history for all rollouts

        Returns
        -------
        k_best_rewards: np.array
            K-rewards of the action trajectories that the highest reward value
        elite_actions: np.array
            Best action histories
        """
        assert rewards.shape == (self.actor.n_planner, 1)
        idxs = np.argsort(rewards, axis=0)

        elite_actions = action_hist[idxs][-self.k_best:, :].squeeze(1)  # sorted (elite, horizon x action_dims)
        k_best_rewards = rewards[idxs][-self.k_best:, :].squeeze(-1)

        assert k_best_rewards.shape == (self.k_best, 1)
        assert elite_actions.shape == (self.k_best, self.actor.horizon * self.actor.action_dims)
        return k_best_rewards, elite_actions

    def update_gaussians(self, old_mu, old_var, best_actions):
        """Updates the mu and var value for the gaussian action sampling method.

        Parameters
        ----------
        old_mu: np.array
            Old mean value
        old_var: np.array
            Old variance value
        best_actions: np.array
            Action history that led to the highest reward

        Returns
        -------
        mu: np.array
            Updated mean values
        var: np.array
            Updated variance values

        """
        assert best_actions.shape == (self.k_best, self.actor.horizon * self.actor.action_dims)

        new_mu = best_actions.mean(0)
        new_var = best_actions.var(0)

        # Softupdate
        mu = (self.update_alpha * old_mu + (1.0 - self.update_alpha) * new_mu)
        var = (self.update_alpha * old_var + (1.0 - self.update_alpha) * new_var)
        assert mu.shape == (self.actor.horizon * self.actor.action_dims,)
        assert var.shape == (self.actor.horizon * self.actor.action_dims,)
        return mu, var

    def acting_step(self, obs, rhs, done, deterministic=False):
        """
        Does the MPC search with CEM action planning process.

        Parameters
        ----------
        obs: torch.tensor
            Current world observation
        rhs: dict
            RNN recurrent hidden states.
        done: torch.tensor
            1.0 if current obs is the last one in the episode, else 0.0.
        deterministic: bool
            Whether to randomly sample action from predicted distribution or taking the mode.

        Returns
        -------
        action: torch.tensor
            Predicted next action.
        clipped_action: torch.tensor
            Predicted next action (clipped to be within action space).
        rhs: batch
            Actor recurrent hidden state.
        other: dict
            Additional MPC predictions, which are not used in other algorithms.
        """

        with torch.no_grad():

            initial_state = obs.repeat(self.actor.n_planner, 1).to(self.device)
            mu = np.zeros(self.actor.horizon * self.actor.action_dims)
            var = 5 * np.ones(self.actor.horizon * self.actor.action_dims)
            X = stats.truncnorm(self.lb, self.ub, loc=np.zeros_like(mu), scale=np.ones_like(mu))
            i = 0

            while i < self.iter_update_steps and np.max(var) > self.epsilon:
                states = initial_state
                returns = np.zeros((self.actor.n_planner, 1))
                # variables
                lb_dist = mu - self.lb
                ub_dist = self.ub - mu
                constrained_var = np.minimum(np.minimum(np.square(lb_dist / 2), np.square(ub_dist / 2)), var)

                actions = X.rvs(size=[self.actor.n_planner, self.actor.horizon * self.actor.action_dims]) * np.sqrt(
                    constrained_var) + mu
                actions = np.clip(actions, self.lb, self.ub)
                actions_t = torch.from_numpy(actions.reshape(
                    self.actor.n_planner, self.actor.horizon, self.actor.action_dims)).float().to(self.device)

                for t in range(self.actor.horizon):
                    with torch.no_grad():
                        states, rewards = self.actor.dynamics_model.predict(states, actions_t[:, t, :])
                    returns += rewards.cpu().numpy()

                k_best_rewards, k_best_actions = self.select_k_best(returns, actions)
                mu, var = self.update_gaussians(mu, var, k_best_actions)
                i += 1

            best_action_sequence = mu.reshape(self.actor.horizon, -1)
            best_action = np.copy(best_action_sequence[0])
            assert best_action.shape == (self.actor.action_dims,)
            action = torch.from_numpy(best_action).float().to(self.device)
            clipped_action = action

        return action.unsqueeze(0), clipped_action.unsqueeze(0), rhs, {}

    def training_step(self, batch):
        """Does the forward pass and loss calculation of the dynamics model given the training data.

        Parameters
        ----------
        batch: dict
            Training data with inputs and labels

        Returns
        -------
            torch.Tensor: Returns the training loss
        """
        train_inputs = batch["train_input"]
        train_labels = batch["train_label"]

        self.actor.train()
        prediction = self.actor.dynamics_model.model(train_inputs)
        loss = self.loss_func(prediction, train_labels)
        return loss

    def compute_gradients(self, batch, grads_to_cpu=True):
        """
        Compute loss and compute gradients but don't do optimization step,
        return gradients instead.

        Parameters
        ----------
        batch: dict
            data batch containing all required tensors to compute dynamics model losses.
        grads_to_cpu: bool
            If gradient tensor will be sent to another node, need to be in CPU.

        Returns
        -------
        grads: list of tensors
            List of actor_critic gradients.
        info: dict
            Dict containing current dynamics model iteration information.
        """
        if batch["batch_number"] == 0:

            # TODO: add reinitialization
            # reinitializes model for new training
            # if self.iter != 0 and self.mb_train_epochs == 0:
            #     self.actor.reinitialize_dynamics_model()
            #     self.actor.to(self.device)
            #     self.dynamics_optimizer = optim.Adam(self.actor.dynamics_model.parameters(), lr=self.lr)

            self.reuse_data = True
            self.mb_train_epochs += 1

        if self.mb_train_epochs == self.num_epochs:
            self.reuse_data = False
            self.mb_train_epochs = 0

        train_loss = self.training_step(batch)

        self.dynamics_optimizer.zero_grad()
        train_loss.backward()
        nn.utils.clip_grad_norm_(self.actor.dynamics_model.parameters(), self.max_grad_norm)
        dyna_grads = get_gradients(self.actor.dynamics_model, grads_to_cpu=grads_to_cpu)

        info = {"train_loss": train_loss.item()}
        grads = {"dyna_grads": dyna_grads}

        # once break condition is used set reuse_data to False
        return grads, info

    def apply_gradients(self, gradients=None):
        """
        Take an optimization step, previously setting new gradients if provided.
        Update also target networks.

        Parameters
        ----------
        gradients : list of tensors
            List of actor gradients.
        """
        if gradients is not None:
            set_gradients(
                self.actor.dynamics_model,
                gradients=gradients["dyna_grads"], device=self.device)

        self.dynamics_optimizer.step()
        self.iter += 1

    def set_weights(self, actor_weights):
        """
        Update actor with the given weights. Update also target networks.

        Parameters
        ----------
        actor_weights : dict of tensors
            Dict containing actor weights to be set.
        """
        self.actor.load_state_dict(actor_weights)
        self.iter += 1

    def update_algorithm_parameter(self, parameter_name, new_parameter_value):
        """
        If `parameter_name` is an attribute of the algorithm, change its value
        to `new_parameter_value value`.

        Parameters
        ----------
        parameter_name : str
            Worker.algo attribute name
        new_parameter_value : int or float
            New value for `parameter_name`.
        """
        if hasattr(self, parameter_name):
            setattr(self, parameter_name, new_parameter_value)
        if parameter_name == "lr":
            for param_group in self.dynamics_optimizer.param_groups:
                param_group['lr'] = new_parameter_value



import torch
import itertools
import torch.nn as nn
import torch.optim as optim

import pytorchrl as prl
from pytorchrl.agent.algorithms.base import Algorithm
from pytorchrl.agent.algorithms.policy_loss_addons import PolicyLossAddOn
from pytorchrl.agent.algorithms.utils import get_gradients, set_gradients


class A2C(Algorithm):
    """
    Algorithm class to execute A2C, from Mnih et al. 2016 (https://arxiv.org/pdf/1602.01783.pdf).

    Parameters
    ----------
    device : torch.device
        CPU or specific GPU where class computations will take place.
    envs : VecEnv
        Vector of environments instance.
    actor : Actor
        Actor_critic class instance.
    lr_v : float
        Value network learning rate.
    lr_pi : float
        Policy network learning rate.
    gamma : float
        Discount factor parameter.
    num_test_episodes : int
        Number of episodes to complete in each test phase.
    max_grad_norm : float
        Gradient clipping parameter.
    test_every : int
        Regularity of test evaluations in actor updates.
    num_test_episodes : int
        Number of episodes to complete in each test phase.
    policy_loss_addons : list
        List of PolicyLossAddOn components adding loss terms to the algorithm policy loss.
    """

    def __init__(self,
                 device,
                 envs,
                 actor,
                 lr_v=1e-4,
                 lr_pi=1e-4,
                 gamma=0.99,
                 test_every=5000,
                 max_grad_norm=0.5,
                 num_test_episodes=5,
                 policy_loss_addons=[]):

        # ---- General algo attributes ----------------------------------------

        # Discount factor
        self._gamma = gamma

        # Number of steps collected with initial random policy
        self._start_steps = int(0)  # Default to 0 for On-policy algos

        # Times data in the buffer is re-used before data collection proceeds
        self._num_epochs = int(1)

        # Number of data samples collected between network update stages
        self._update_every = None  # Depends on storage capacity

        # Number mini batches per epoch
        self._num_mini_batch = int(1)

        # Size of update mini batches
        self._mini_batch_size = None  # Depends on storage capacity

        # Number of network updates between test evaluations
        self._test_every = int(test_every)

        # Number of episodes to complete when testing
        self._num_test_episodes = int(num_test_episodes)

        # ---- A2C-specific attributes ----------------------------------------

        self.iter = 0
        self.envs = envs
        self.actor = actor
        self.device = device
        self.max_grad_norm = max_grad_norm

        assert hasattr(self.actor, "value_net1"), "A2C requires value critic (num_critics=1)"

        # ----- Policy Loss Addons --------------------------------------------

        # Sanity check, policy_loss_addons is a PolicyLossAddOn instance
        # or a list of PolicyLossAddOn instances
        assert isinstance(policy_loss_addons, (PolicyLossAddOn, list)),\
            "A2C policy_loss_addons parameter should be a  PolicyLossAddOn instance " \
            "or a list of PolicyLossAddOn instances"
        if isinstance(policy_loss_addons, list):
            for addon in policy_loss_addons:
                assert isinstance(addon, PolicyLossAddOn), \
                    "A2C policy_loss_addons parameter should be a  PolicyLossAddOn" \
                    " instance or a list of PolicyLossAddOn instances"
        else:
            policy_loss_addons = [policy_loss_addons]

        self.policy_loss_addons = policy_loss_addons
        for addon in self.policy_loss_addons:
            addon.setup(self.actor, self.device)

        # ----- Optimizer -----------------------------------------------------

        self.pi_optimizer = optim.Adam(self.actor.policy_net.parameters(), lr=lr_pi)
        self.v_optimizer = optim.Adam(self.actor.value_net1.parameters(), lr=lr_v)

    @classmethod
    def create_factory(cls,
                       lr_v=1e-4,
                       lr_pi=1e-4,
                       gamma=0.99,
                       test_every=5000,
                       max_grad_norm=0.5,
                       num_test_episodes=5,
                       policy_loss_addons=[]):
        """
        Returns a function to create new A2C instances.

        Parameters
        ----------
        lr_v : float
            Value network learning rate.
        lr_pi : float
            Policy network learning rate.
        gamma : float
            Discount factor parameter.
        num_test_episodes : int
            Number of episodes to complete in each test phase.
        max_grad_norm : float
            Gradient clipping parameter.
        test_every : int
            Regularity of test evaluations in actor updates.
        num_test_episodes : int
            Number of episodes to complete in each test phase.
        policy_loss_addons : list
            List of PolicyLossAddOn components adding loss terms to the algorithm policy loss.

        Returns
        -------
        create_algo_instance : func
            Function that creates a new A2C class instance.
        algo_name : str
            Name of the algorithm.
        """

        def create_algo_instance(device, actor, envs):
            return cls(lr_pi=lr_pi,
                       lr_v=lr_v,
                       envs=envs,
                       actor=actor,
                       gamma=gamma,
                       device=device,
                       test_every=test_every,
                       max_grad_norm=max_grad_norm,
                       num_test_episodes=num_test_episodes,
                       policy_loss_addons=policy_loss_addons)

        return create_algo_instance, prl.A2C

    @property
    def gamma(self):
        """Returns discount factor gamma."""
        return self._gamma

    @property
    def start_steps(self):
        """Returns the number of steps to collect with initial random policy."""
        return self._start_steps

    @property
    def num_epochs(self):
        """
        Returns the number of times the whole buffer is re-used before data
        collection proceeds.
        """
        return self._num_epochs

    @property
    def update_every(self):
        """
        Returns the number of data samples collected between
        network update stages.
        """
        return self._update_every

    @property
    def num_mini_batch(self):
        """
        Returns the number of times the whole buffer is re-used before data
        collection proceeds.
        """
        return self._num_mini_batch

    @property
    def mini_batch_size(self):
        """
        Returns the number of mini batches per epoch.
        """
        return self._mini_batch_size

    @property
    def test_every(self):
        """Number of network updates between test evaluations."""
        return self._test_every

    @property
    def num_test_episodes(self):
        """
        Returns the number of episodes to complete when testing.
        """
        return self._num_test_episodes

    def acting_step(self, obs, rhs, done, deterministic=False):
        """
        A2C acting function.

        Parameters
        ----------
        obs: torch.tensor
            Current world observation
        rhs: torch.tensor
            RNN recurrent hidden state (if policy is not a RNN, rhs will contain zeroes).
        done: torch.tensor
            1.0 if current obs is the last one in the episode, else 0.0.
        deterministic: bool
            Whether to randomly sample action from predicted distribution or take the mode.

        Returns
        -------
        action: torch.tensor
            Predicted next action.
        clipped_action: torch.tensor
            Predicted next action (clipped to be within action space).
        rhs: torch.tensor
            Policy recurrent hidden state (if policy is not a RNN, rhs will contain zeroes).
        other: dict
            Additional A2C predictions, value score and action log probability.
        """

        with torch.no_grad():

            (action, clipped_action, logp_action, rhs,
             entropy_dist, dist) = self.actor.get_action(
                obs, rhs, done, deterministic)

            value_dict = self.actor.get_value(obs, rhs, done)
            value = value_dict.get("value_net1")
            rhs = value_dict.get("rhs")

            other = {prl.VAL: value, prl.LOGP: logp_action}

        return action, clipped_action, rhs, other

    def compute_loss(self, data):
        """
        Calculate A2C loss

        Parameters
        ----------
        data: dict
            Data batch dict containing all required tensors to compute A2C loss.

        Returns
        -------
        loss : torch.tensor
            A2C loss.
        """

        o, rhs, a, old_v = data[prl.OBS], data[prl.RHS], data[prl.ACT], data[prl.VAL]
        r, d, old_logp, adv = data[prl.RET], data[prl.DONE], data[prl.LOGP], data[prl.ADV]

        # Policy loss
        logp, dist_entropy, dist = self.actor.evaluate_actions(o, rhs, d, a)
        pi_loss = - (logp * adv).mean()

        # Extend policy loss with addons
        addons_info = {}
        for addon in self.policy_loss_addons:
            addon_loss, addons_info = addon.compute_loss_term(data, dist, addons_info)
            pi_loss += addon_loss

        # Value loss
        new_v = self.actor.get_value(o, rhs, d).get("value_net1")
        value_loss = (r - new_v).pow(2).mean()

        return pi_loss, value_loss, addons_info

    def compute_gradients(self, batch, grads_to_cpu=True):
        """
        Compute loss and compute gradients but don't do optimization step,
        return gradients instead.

        Parameters
        ----------
        data: dict
            data batch containing all required tensors to compute A2C loss.
        grads_to_cpu: bool
            If gradient tensor will be sent to another node, need to be in CPU.

        Returns
        -------
        grads: list of tensors
            List of actor gradients.
        info: dict
            Dict containing current A2C iteration information.
        """

        # Compute A2C losses
        action_loss, value_loss, addons_info = self.compute_loss(batch)

        # Compute policy gradients
        self.pi_optimizer.zero_grad()
        action_loss.backward(retain_graph=True)

        for p in self.actor.policy_net.parameters():
            p.requires_grad = False

        # Compute value gradients
        self.v_optimizer.zero_grad()
        value_loss.backward()

        for p in self.actor.policy_net.parameters():
            p.requires_grad = True

        # Clip gradients to max value
        nn.utils.clip_grad_norm_(self.actor.parameters(), self.max_grad_norm)

        pi_grads = get_gradients(self.actor.policy_net, grads_to_cpu=grads_to_cpu)
        v_grads = get_gradients(self.actor.value_net1, grads_to_cpu=grads_to_cpu)
        grads = {"pi_grads": pi_grads, "v_grads": v_grads}

        info = {
            "value_loss": value_loss.item(),
            "action_loss": action_loss.item(),
        }

        info.update(addons_info)

        return grads, info

    def apply_gradients(self, gradients=None):
        """
        Take an optimization step, previously setting new gradients if provided.

        Parameters
        ----------
        gradients: list of tensors
            List of actor gradients.
        """
        if gradients is not None:
            set_gradients(
                self.actor.policy_net,
                gradients=gradients["pi_grads"], device=self.device)
            set_gradients(
                self.actor.value_net1,
                gradients=gradients["v_grads"], device=self.device)

        self.pi_optimizer.step()
        self.v_optimizer.step()
        self.iter += 1

    def set_weights(self, actor_weights):
        """
        Update actor with the given weights.

        Parameters
        ----------
        actor_weights: dict of tensors
            Dict containing actor weights to be set.
        """
        self.actor.load_state_dict(actor_weights)
        self.iter += 1

    def update_algorithm_parameter(self, parameter_name, new_parameter_value):
        """
        If `parameter_name` is an attribute of the algorithm, change its value
        to `new_parameter_value value`.

        Parameters
        ----------
        parameter_name : str
            Worker.algo attribute name
        new_parameter_value : int or float
            New value for `parameter_name`.
        """
        if hasattr(self, parameter_name):
            setattr(self, parameter_name, new_parameter_value)
        if parameter_name == "lr_v":
            for param_group in self.v_optimizer.param_groups:
                param_group['lr'] = new_parameter_value
        elif parameter_name == "lr_pi":
            for param_group in self.pi_optimizer.param_groups:
                param_group['lr'] = new_parameter_value



"""
Implementation of the decorator using a Encoder-Decoder architecture.
Adapted from https://github.com/MolecularAI/reinvent-models.
"""
import math
import torch
import torch.nn as nn
import torch.nn.utils.rnn as tnnur
from pytorchrl.agent.actors.feature_extractors import Embedding


class Encoder(nn.Module):
    """LSTM Bidirectional RNN encoder model."""

    def __init__(self, vocabulary_size, num_layers=3, num_dimensions=512, dropout=0.0):
        """
        Initializes the LSTM Bidirectional RNN encoder model.

        Parameters
        ----------
        vocabulary_size : int
            Number of possible tokens in the input space.
        num_dimensions : int
            Recurrent hidden state and output size.
        dropout : float
            If non-zero, introduces a Dropout layer on the outputs of each GRU layer except the last layer.
        num_layers : int
            Number of recurrent layers.
        """

        super(Encoder, self).__init__()

        self.dropout = dropout
        self.num_layers = num_layers
        self.num_dimensions = num_dimensions
        self.vocabulary_size = vocabulary_size

        self._embedding = nn.Sequential(
            nn.Embedding(self.vocabulary_size, self.num_dimensions),
            nn.Dropout(dropout))

        self._rnn = nn.LSTM(
            self.num_dimensions, self.num_dimensions, self.num_layers,
            batch_first=True, dropout=self.dropout, bidirectional=True)

    def forward(self, padded_seqs, seq_lengths):
        """
        Performs encoder forward pass.

        Parameters
        ----------
        padded_seqs : torch.tensor
            A tensor with the sequences (batch, seq).
        seq_lengths : torch.tensor
            1D tensor with the lengths of the sequences (for packed sequences).

        Returns
        -------
        padded_seqs : torch.tensor
            LSTM outputs values.
        (hs_h, hs_c) : tuple
            LSTM last recurrent hidden state.
        """

        batch_size = padded_seqs.size(0)
        hidden_state = self._initialize_hidden_state(batch_size)

        padded_seqs = self._embedding(padded_seqs.long())
        hs_h, hs_c = (hidden_state, hidden_state.clone().detach())

        packed_seqs = tnnur.pack_padded_sequence(padded_seqs, seq_lengths, batch_first=True, enforce_sorted=False)
        packed_seqs, (hs_h, hs_c) = self._rnn(packed_seqs, (hs_h, hs_c))
        padded_seqs, _ = tnnur.pad_packed_sequence(packed_seqs, batch_first=True)

        # sum up bidirectional layers and collapse
        hs_h = hs_h.view(self.num_layers, 2, batch_size, self.num_dimensions).sum(dim=1)
        hs_c = hs_c.view(self.num_layers, 2, batch_size, self.num_dimensions).sum(dim=1)
        padded_seqs = padded_seqs.view(batch_size, -1, 2, self.num_dimensions).sum(dim=2).squeeze(2)  # (batch, seq, dim)

        return padded_seqs, (hs_h, hs_c)

    def _initialize_hidden_state(self, batch_size):
        """Returns a tensor of zeros with the expected shape of the model's rhs."""
        return torch.zeros(self.num_layers*2, batch_size, self.num_dimensions).cuda()


class AttentionLayer(nn.Module):

    def __init__(self, num_dimensions):
        super(AttentionLayer, self).__init__()

        self.num_dimensions = num_dimensions

        self._attention_linear = nn.Sequential(
            nn.Linear(self.num_dimensions*2, self.num_dimensions), nn.Tanh())

    def forward(self, padded_seqs, encoder_padded_seqs, decoder_mask):
        """Performs the forward pass.

        Parameters
        ----------
        padded_seqs : torch.tensor
            A tensor with the decoder output sequences (batch, seq_d, dim).
        encoder_padded_seqs : torch.tensor
             tensor with the encoded context sequences (batch, seq_e, dim).
        decoder_mask : torch.tensor
            A tensor that represents the encoded input mask.

        Returns
        -------
        logits : torch.tensor
            Modified logits.
        attention_weights : torch.tensor
            Model attention weights.
        """

        # scaled dot-product
        # (batch, seq_d, 1, dim) * (batch, 1, seq_e, dim) => (batch, seq_d, seq_e)
        attention_weights = (padded_seqs.unsqueeze(dim=2) * encoder_padded_seqs.unsqueeze(dim=1)).sum(
            dim=3).div(math.sqrt(self.num_dimensions)).softmax(dim=2)

        # (batch, seq_d, seq_e) @ (batch, seq_e, dim) => (batch, seq_d, dim)
        attention_context = attention_weights.bmm(encoder_padded_seqs)
        return (self._attention_linear(torch.cat(
            [padded_seqs, attention_context], dim=2)) * decoder_mask, attention_weights)


class Decoder(nn.Module):
    """LSTM RNN decoder model."""

    def __init__(self, vocabulary_size, num_layers=3, num_dimensions=512, dropout=0.0):
        """
        Initializes the LSTM decoder model.

        Parameters
        ----------
        vocabulary_size : int
            Number of possible tokens in the input space.
        num_dimensions : int
            Recurrent hidden state and output size.
        dropout : float
            If non-zero, introduces a Dropout layer on the outputs of each GRU layer except the last layer.
        num_layers : int
            Number of recurrent layers.
        """
        super(Decoder, self).__init__()

        self.num_layers = num_layers
        self.num_dimensions = num_dimensions
        self.vocabulary_size = vocabulary_size
        self.dropout = dropout

        self._embedding = nn.Sequential(
            nn.Embedding(self.vocabulary_size, self.num_dimensions),
            nn.Dropout(dropout))

        self._rnn = nn.LSTM(
            self.num_dimensions, self.num_dimensions, self.num_layers,
            batch_first=True, dropout=self.dropout, bidirectional=False)

        self._attention = AttentionLayer(self.num_dimensions)

    def forward(self, padded_seqs, seq_lengths, encoder_padded_seqs, hidden_states):
        """
        Performs decoder forward pass.

        Parameters
        ----------
        padded_seqs : torch.tensor
            A tensor with the sequences (batch, seq).
        seq_lengths : torch.tensor
            1D tensor with the lengths of the sequences (for packed sequences).
        encoder_padded_seqs : torch.tensor
             tensor with the encoded context sequences (batch, seq_e, dim).
        hidden_states : tuple
            Last recurrent hidden states.

        Returns
        -------
        logits : torch.tensor
            Output logits.
        hidden_states : tuple
            Updated recurrent hidden state.
        attention_weights : tensor
            Attention layer weights.
        """

        padded_encoded_seqs = self._embedding(padded_seqs.long())

        packed_encoded_seqs = tnnur.pack_padded_sequence(padded_encoded_seqs, seq_lengths, batch_first=True, enforce_sorted=False)
        packed_encoded_seqs, hidden_states = self._rnn(packed_encoded_seqs, hidden_states)
        padded_encoded_seqs, _ = tnnur.pad_packed_sequence(packed_encoded_seqs, batch_first=True)  # (batch, seq, dim)

        mask = (padded_encoded_seqs[:, :, 0] != 0).unsqueeze(dim=-1).type(torch.float)
        attn_padded_encoded_seqs, attention_weights = self._attention(padded_encoded_seqs, encoder_padded_seqs, mask)
        logits = attn_padded_encoded_seqs * mask

        return logits, hidden_states, attention_weights


class LSTMEncoderDecoder(nn.Module):
    """
    An encoder-decoder model that encodes context and then uses a decoder and an attention layer
    to autoregressively generate actions.

    Action space is expected to be a dict, containing at least the following fields:
        - context: padded context sequences, expected shape=(batch_size, sequences_length)
        - context_length: a 1D tensor with the length of each context sequence.
        - obs: next episode observations, shape=(batch_size, **obs_input_shape)
        - obs_length: a 1D tensor with the length of each obs sequence (can be 1).

    Parameters
    ----------
    input_size : int
        Input feature map size (unused).
    encoder_params : dict
        Dictionary specifying encoder's vocabulary_size, num_layers (default 3), num_dimensions (default 512) and
        dropout (default 0.0).
    decoder_params : dict
        Dictionary specifying decoder's vocabulary_size, num_layers (default 3), num_dimensions (default 512) and
        dropout (default 0.0).
    """

    def __init__(self, input_size, encoder_params={}, decoder_params={}):
        super(LSTMEncoderDecoder, self).__init__()

        self._encoder = Encoder(**encoder_params)
        self._decoder = Decoder(**decoder_params)
        self.encoder_rhs = None
        self.encoder_padded_seqs = None

    def _forward_encoder(self, padded_seqs, seq_lengths):
        """
        Does a forward pass only of the encoder.

        Parameters
        ----------
        padded_seqs : torch.tensor
            Input data to feed the encoder.
        seq_lengths : torch.tensor
            1D tensor with the lengths of each padded_seqs sequences (for packed sequences).

        Returns
        -------
        encoded_seqs : torch.tensor
            Output encoded sequences.
        hidden_states : tuple
            Updated recurrent hidden state.
        """
        return self._encoder(padded_seqs, seq_lengths)

    def _forward_decoder(self, padded_seqs, seq_lengths, encoder_padded_seqs, hidden_states):
        """
        Does a forward pass only of the decoder.

        Parameters
        ----------
        padded_seqs : torch.tensor
            Input data to feed the decoder.
        seq_lengths : torch.tensor
            1D tensor with the lengths of each padded_seqs sequences (for packed sequences).
        hidden_states : tuple
            Last recurrent hidden states.

        Returns
        -------
        logits : torch.tensor
            Output decoder logits.
        hidden_states : tuple
            Updated recurrent hidden state.
        """
        return self._decoder(padded_seqs, seq_lengths, encoder_padded_seqs, hidden_states)

    def _forward_encoder_decoder(self, encoder_seqs, encoder_seq_lengths, decoder_seqs, decoder_seq_lengths, hxs, done):

        encoder_seq_lengths = encoder_seq_lengths.cpu().long()
        decoder_seq_lengths = decoder_seq_lengths.cpu().long()
        masks = 1 - done

        if decoder_seqs.size(0) == hxs.size(0):

            if self.encoder_rhs is None or self.encoder_padded_seqs is None:
                self.encoder_padded_seqs, self.encoder_rhs = self._forward_encoder(encoder_seqs, encoder_seq_lengths)
                self.encoder_rhs = torch.transpose(torch.cat(self.encoder_rhs), 0, 1)

            # Replace "done" hxs by self.encoder_rhs
            hxs = torch.where(done.unsqueeze(-1) > 0.0, self.encoder_rhs, hxs)

            # Chunk rhs (where does this go?)
            hxs = torch.chunk((torch.transpose(hxs, 0, 1)), 2)

            logits, hxs, _ = self._forward_decoder(decoder_seqs, decoder_seq_lengths, self.encoder_padded_seqs, hxs)
            logits = logits.squeeze(1)

            hxs = torch.transpose(torch.cat(hxs), 0, 1)

        else:

            # x is a (T, N, -1) tensor that has been flatten to (T * N, -1)
            N = hxs.size(0)
            T = int(decoder_seqs.size(0) / N)

            # Set encoder outputs to None
            self.encoder_rhs = None
            self.encoder_padded_seqs = None

            # unflatten
            decoder_seqs = torch.transpose(decoder_seqs.view(T, N, -1), 0, 1)
            encoder_seqs = encoder_seqs.view(T, N, -1)[0]
            encoder_seq_lengths[encoder_seq_lengths == 0] = encoder_seqs.size(1)
            encoder_seq_lengths = encoder_seq_lengths.view(T, N)[0]

            # Same deal with masks
            masks = torch.transpose(masks.view(T, N), 0, 1)

            # Let's figure out which steps in the sequence have a zero for any agent
            # We will always assume t=0 has a zero in it as that makes the logic cleaner
            has_zeros = torch.nonzero(((masks[:, 1:] == 0.0).any(dim=0)), as_tuple=False).squeeze().cpu()

            # +1 to correct the masks[1:]
            if has_zeros.dim() == 0:
                # Deal with scalar
                has_zeros = [has_zeros.item() + 1]
            else:
                has_zeros = (has_zeros + 1).numpy().tolist()

            # add t=0 and t=T to the list
            has_zeros = [0] + has_zeros + [T]

            outputs = []
            for i in range(len(has_zeros) - 1):
                # We can now process steps that don't have any zeros in masks together!
                # This is much faster
                start_idx = has_zeros[i]
                end_idx = has_zeros[i + 1]

                # TODO: run encoder in position start_idx if required
                if self.encoder_rhs is None or self.encoder_padded_seqs is None:
                    self.encoder_padded_seqs, self.encoder_rhs = self._forward_encoder(
                        encoder_seqs, encoder_seq_lengths)
                    self.encoder_rhs = torch.transpose(torch.cat(self.encoder_rhs), 0, 1)

                hxs = torch.where(masks[:, start_idx: start_idx + 1].unsqueeze(-1) == 0.0, self.encoder_rhs, hxs)

                # Chunk rhs (where does this go?)
                hxs = torch.chunk((torch.transpose(hxs, 0, 1)), 2)

                # Run decoder from start_idx to end_idx
                lengths = torch.cat([torch.LongTensor([end_idx - start_idx])] * N)
                logits, hxs, _ = self._forward_decoder(
                    decoder_seqs[:, start_idx:end_idx].squeeze(-1), lengths, self.encoder_padded_seqs, hxs)

                hxs = torch.transpose(torch.cat(hxs), 0, 1)

                outputs.append(logits)

            # x is a (T, N, -1) tensor
            logits = torch.cat(outputs, dim=1)

            # flatten
            logits = torch.transpose(logits, 0, 1).reshape(T * N, -1)

            # Set encoder outputs to None
            self.encoder_rhs = None
            self.encoder_padded_seqs = None

        return logits, hxs

    def forward(self, inputs, rhs, done):
        """
        Forward pass Neural Network

        Parameters
        ----------
        inputs : torch.tensor
            A tensor containing episode observations.
        rhs : torch.tensor
            A tensor representing the recurrent hidden states.
        done : torch.tensor
            A tensor indicating where episodes end.

        Returns
        -------
        x : torch.tensor
            Output feature map.
        rhs : torch.tensor
            Updated recurrent hidden state.
        """

        encoder_seqs = inputs["context"]
        decoder_seqs = inputs["obs"]
        encoder_seq_lengths = inputs["context_length"].cpu().long()
        decoder_seq_lengths = inputs["obs_length"].cpu().long()

        logits, rhs = self._forward_encoder_decoder(
            encoder_seqs, encoder_seq_lengths, decoder_seqs, decoder_seq_lengths, rhs, done)

        return logits, rhs

    @property
    def num_outputs(self):
        """Output feature map size (as in np.prod(self.output_shape))."""
        return self._decoder.num_dimensions

    @property
    def recurrent_hidden_state_size(self):
        """Recurrent hidden state size"""
        return self._decoder.num_dimensions

    def get_initial_recurrent_state(self, num_proc):
        """Returns a tensor of zeros with the expected shape of the model's rhs."""
        return torch.zeros(num_proc, self._encoder.num_layers * 2, self._encoder.num_dimensions)



import gym
from pytorchrl.envs.common import FrameStack, FrameSkip, DelayedReward, CartPoleActionWrapper


def mujoco_train_env_factory(env_id, index_col_worker, index_grad_worker, index_env=0, seed=0, frame_skip=0, frame_stack=1, reward_delay=1):
    """
    Create train MuJoCo environment.

    Parameters
    ----------
    env_id : str
        Environment name.
    index_col_worker : int
        Index of the collection worker running this environment.
    index_grad_worker : int
        Index of the gradient worker running the collection worker running this environment.
    index_env : int
        Index of this environment withing the vector of environments.
    seed : int
        Environment random seed.
    frame_skip : int
        Return only every `frame_skip`-th observation.
    frame_stack : int
        Observations composed of last `frame_stack` frames stacked.
    reward_delay : int
        Only return accumulated reward every `reward_delay` steps to simulate sparse reward environment.

    Returns
    -------
    env : gym.Env
        Train environment.
    """

    env = gym.make(env_id)
    env.seed(index_grad_worker * 1000 + 100 * index_col_worker + index_env + seed)

    if frame_skip > 0:
        env = FrameSkip(env, skip=frame_skip)

    if frame_stack > 1:
        env = FrameStack(env, k=frame_stack)

    if reward_delay > 1:
        env = DelayedReward(env, delay=reward_delay)
    
    if env_id == "CartPole-v0" or env_id == "CartPole-v1":
        env = CartPoleActionWrapper(env)

    return env


def mujoco_test_env_factory(env_id, index_col_worker, index_grad_worker, index_env=0, seed=0, frame_skip=0, frame_stack=1, reward_delay=1):
    """
    Create test MuJoCo environment.

    Parameters
    ----------
    env_id : str
        Environment name.
    index_col_worker : int
        Index of the collection worker running this environment.
    index_grad_worker : int
        Index of the gradient worker running the collection worker running this environment.
    index_env : int
        Index of this environment withing the vector of environments.
    seed : int
        Environment random seed.
    frame_skip : int
        Return only every `frame_skip`-th observation.
    frame_stack : int
        Observations composed of last `frame_stack` frames stacked.
    reward_delay : int
        Only return accumulated reward every `reward_delay` steps to simulate sparse reward environment.

    Returns
    -------
    env : gym.Env
        Test environment.
    """
    env = gym.make(env_id)
    env.seed(index_grad_worker * 1000 + 100 * index_col_worker + index_env + seed)

    if frame_skip > 0:
        env = FrameSkip(env, skip=frame_skip)

    if frame_stack > 1:
        env = FrameStack(env, k=frame_stack)

    if reward_delay > 1:
        env = DelayedReward(env, delay=reward_delay)

    if env_id == "CartPole-v0" or env_id == "CartPole-v1":
        env = CartPoleActionWrapper(env)

    return env



import gym
import gymnasium
from gymnasium.core import Wrapper


class RewardShapeWrapper(Wrapper):
    """Wrapper to change certain reward values at specific positions."""

    def __init__(self, env, positions=[], scale=1.0):
        self.target_positions = positions  # [(env.width - 2, env.height - 2)]
        self.scale = scale
        super().__init__(env)

    def step(self, action):
        obs, reward, terminated, truncated, info = self.env.step(action)
        if self.env.agent_pos in self.target_positions:
            reward *= self.scale
        return obs, reward, terminated, truncated, info


class Gymnasium2GymWrapper(Wrapper):
    """Wrapper to translate the Gymnasium interface into Gym interface."""

    def __init__(self, env):
        super().__init__(env)

        # Translate action space
        if isinstance(self.env.action_space, gymnasium.spaces.Discrete):
            self.env.action_space = gym.spaces.Discrete(self.env.action_space.n)
        elif isinstance(self.env.action_space, gymnasium.spaces.Box):
            self.env.action_space = gym.spaces.Box(
                high=self.env.action_space.high,
                low=self.env.action_space.low,
                shape=self.env.action_space.shape,
                dtype=self.env.action_space.dtype)

        # Translate observation space
        if isinstance(self.env.observation_space, gymnasium.spaces.Discrete):
            self.env.observation_space = gym.spaces.Discrete(self.env.observation_space.n)
        elif isinstance(self.env.observation_space, gymnasium.spaces.Box):
            self.env.observation_space = gym.spaces.Box(
                high=self.env.observation_space.high,
                low=self.env.observation_space.low,
                shape=self.env.observation_space.shape,
                dtype=self.env.observation_space.dtype)

    def step(self, action):
        obs, reward, terminated, truncated, info = self.env.step(action)
        return obs, reward, max(terminated, truncated), info

    def reset(self, **kwargs):
        obs, _ = self.env.reset(**kwargs)
        return obs




import gym
import numpy as np
import torch
import torch.nn as nn
from collections import OrderedDict

import pytorchrl as prl
from pytorchrl.agent.actors.base import Actor
from pytorchrl.agent.actors.distributions import get_dist
from pytorchrl.agent.actors.utils import Scale, Unscale, init
from pytorchrl.agent.actors.feature_extractors import default_feature_extractor


class OnPolicyActor(Actor):
    """
    Actor critic class for On-Policy algorithms.

    It contains a policy network to predict next actions and a critic
    value network to predict the value score of a given obs.

    Parameters
    ----------
    device: torch.device
        CPU or specific GPU where class computations will take place.
    input_space : gym.Space
        Environment observation space.
    action_space : gym.Space
        Environment action space.
    algorithm_name : str
        Name of the RL algorithm used for learning.
    checkpoint : str
        Path to a previously trained Actor checkpoint to be loaded.
    recurrent_net : bool
        Whether to use a RNNs on top of the feature extractors.
    recurrent_net_kwargs:
        Keyword arguments for the memory network.
    feature_extractor_network : nn.Module
        PyTorch nn.Module used as the features extraction block in all networks.
    feature_extractor_kwargs : dict
        Keyword arguments for the feature extractor network.
    shared_policy_value_network : bool
        Whether or not to share weights between policy and value networks.
    """

    def __init__(self,
                 device,
                 input_space,
                 action_space,
                 algorithm_name,
                 checkpoint=None,
                 recurrent_net=None,
                 recurrent_net_kwargs={},
                 feature_extractor_network=None,
                 feature_extractor_kwargs={},
                 shared_policy_value_network=True):

        super(OnPolicyActor, self).__init__(
            device=device,
            checkpoint=checkpoint,
            input_space=input_space,
            action_space=action_space)

        self.recurrent_net = recurrent_net
        self.recurrent_net_kwargs = recurrent_net_kwargs
        self.feature_extractor_network = feature_extractor_network
        self.shared_policy_value_network = shared_policy_value_network
        self.feature_extractor_kwargs = feature_extractor_kwargs

        if algorithm_name in (prl.A2C, prl.PPO):
            self.num_critics_ext = 1
            self.num_critics_int = 0
        elif algorithm_name in (prl.RND_PPO):
            self.num_critics_ext = 1
            self.num_critics_int = 1

        # ----- Policy Network ----------------------------------------------------

        self.create_policy("policy_net")

        # ----- Value Networks ----------------------------------------------------

        for i in range(self.num_critics_ext):
            self.create_critic("value_net{}".format(i + 1))

        for i in range(self.num_critics_int):
            self.create_critic("ivalue_net{}".format(i + 1))

    @classmethod
    def create_factory(
            cls,
            input_space,
            action_space,
            algorithm_name,
            restart_model=None,
            recurrent_net=None,
            recurrent_net_kwargs={},
            feature_extractor_kwargs={},
            feature_extractor_network=None,
            shared_policy_value_network=True):
        """
        Returns a function that creates actor critic instances.

        Parameters
        ----------
        input_space : gym.Space
            Environment observation space.
        action_space : gym.Space
            Environment action space.
        algorithm_name : str
            Name of the RL algorithm_name used for learning.
        restart_model : str
            Path to a previously trained Actor checkpoint to be loaded.
        feature_extractor_network : nn.Module
            PyTorch nn.Module used as the features extraction block in all networks.
        feature_extractor_kwargs : dict
            Keyword arguments for the feature extractor network.
        recurrent_net : nn.Module
             PyTorch nn.Module to use after the feature extractors.
        recurrent_net_kwargs:
            Keyword arguments for the memory network.
        shared_policy_value_network : bool
            Whether or not to share weights between policy and value networks.

        Returns
        -------
        create_actor_critic_instance : func
            creates a new OnPolicyActor class instance.
        """

        def create_actor_critic_instance(device):
            """Create and return an actor critic instance."""
            policy = cls(device=device,
                         input_space=input_space,
                         action_space=action_space,
                         algorithm_name=algorithm_name,
                         recurrent_net=recurrent_net,
                         checkpoint=restart_model,
                         recurrent_net_kwargs=recurrent_net_kwargs,
                         feature_extractor_kwargs=feature_extractor_kwargs,
                         feature_extractor_network=feature_extractor_network,
                         shared_policy_value_network=shared_policy_value_network)
            policy.to(device)

            try:
                policy.try_load_from_checkpoint()
            except RuntimeError:
                pass

            return policy

        return create_actor_critic_instance

    @property
    def is_recurrent(self):
        """Returns True if the actor network are recurrent."""
        return self.recurrent_net

    @property
    def recurrent_hidden_state_size(self):
        """Size of policy recurrent hidden state"""
        return self.recurrent_size

    def actor_initial_states(self, obs):
        """
        Returns all actor inputs required to predict initial action.

        Parameters
        ----------
        obs : torch.tensor
            Initial environment observation.

        Returns
        -------
        obs : torch.tensor
            Initial environment observation.
        rhs : dict
            Initial recurrent hidden states.
        done : torch.tensor
            Initial done tensor, indicating the environment is not done.
        """
        if isinstance(obs, dict):
            num_proc = list(obs.values())[0].size(0)
            dev = list(obs.values())[0].device
        else:
            num_proc = obs.size(0)
            dev = obs.device

        done = torch.zeros(num_proc, 1).to(dev)
        try:
            rhs_policy = self.policy_net.memory_net.get_initial_recurrent_state(num_proc).to(dev)
        except Exception:
            rhs_policy = torch.zeros(num_proc, self.recurrent_hidden_state_size).to(dev)

        rhs = {"policy": rhs_policy}
        rhs.update({"value_net{}".format(i + 1): rhs_policy.clone() for i in range(self.num_critics_ext)})
        rhs.update({"ivalue_net{}".format(i + 1): rhs_policy.clone() for i in range(self.num_critics_int)})

        return obs, rhs, done

    def get_action(self, obs, rhs, done, deterministic=False):
        """
        Predict and return next action, along with other information.

        Parameters
        ----------
        obs : torch.tensor
            Current environment observation.
        rhs : dict
            Current recurrent hidden states.
        done : torch.tensor
            Current done tensor, indicating if episode has finished.
        deterministic : bool
            Whether to randomly sample action from predicted distribution or take the mode.

        Returns
        -------
        action : torch.tensor
            Next action sampled.
        clipped_action : torch.tensor
            Next action sampled, but clipped to be within the env action space.
        logp_action : torch.tensor
            Log probability of `action` within the predicted action distribution.
        rhs : dict
            Updated recurrent hidden states.
        entropy_dist : torch.tensor
            Entropy of the predicted action distribution.
        dist : torch.Distribution
            Predicted probability distribution over next action.
        """

        features = self.policy_net.feature_extractor(obs)
        if self.recurrent_net:
            features, rhs["policy"] = self.policy_net.memory_net(
                features, rhs["policy"], done)
        (action, clipped_action, logp_action, entropy_dist, dist) = self.policy_net.dist(
            features, deterministic=deterministic)

        self.last_action_features = features

        if self.unscale:
            action = self.unscale(action)
            clipped_action = self.unscale(clipped_action)

        return action, clipped_action, logp_action, rhs, entropy_dist, dist

    def evaluate_actions(self, obs, rhs, done, action):
        """
        Evaluate log likelihood of action given obs and the current
        policy network. Returns also entropy distribution.

        Parameters
        ----------
        obs : torch.tensor
            Environment observation.
        rhs : dict
            Recurrent hidden states.
        done : torch.tensor
            Done tensor, indicating if episode has finished.
        action : torch.tensor
            Evaluated action.

        Returns
        -------
        logp_action : torch.tensor
            Log probability of `action` according to the action distribution
            predicted with current version of the policy_net.
        entropy_dist : torch.tensor
            Entropy of the action distribution predicted with current version
            of the policy_net.
        dist : torch.Distribution
            Predicted probability distribution over next action.
        """

        if self.scale:
            action = self.scale(action)

        features = self.policy_net.feature_extractor(obs)

        if self.recurrent_net:
            features, rhs["policy"] = self.policy_net.memory_net(features, rhs["policy"], done)

        logp_action, entropy_dist, dist = self.policy_net.dist.evaluate_pred(features, action)

        self.last_action_features = features

        return logp_action, entropy_dist, dist

    def get_value_specific_net(self, obs, rhs, done, value_net_name):
        """
        Return value score for a single value network.

        Parameters
        ----------
        obs : torch.tensor
            Environment observation.
        rhs : dict
            Recurrent hidden states.
        done : torch.tensor
            Done tensor, indicating if episode has finished.

        Returns
        -------
        value : torch.tensor
            Predicted value score.
        rhs : dict
            Updated recurrent hidden states.
        """

        value_net = getattr(self, value_net_name)

        if self.shared_policy_value_network:
            if self.last_action_features is None or self.last_action_features.shape[0] != done.shape[0]:
                _, _, _, _, _, _ = self.get_action(obs, rhs["policy"], done)
            value = value_net.predictor(self.last_action_features)

        else:
            features = value_net.feature_extractor(obs)
            if self.recurrent_net:
                features, rhs[value_net_name] = value_net.memory_net(
                    features, rhs[value_net_name], done)
            value = value_net.predictor(features)

        return value, rhs

    def get_value(self, obs, rhs, done):
        """
        Return all value scores of given observation.

        Parameters
        ----------
        obs : torch.tensor
            Environment observation.
        rhs : dict
            Recurrent hidden states.
        done : torch.tensor
            Done tensor, indicating if episode has finished.

        Returns
        -------
        output : dict
            Dict containing value prediction from each critic under keys "value_net1", "value_net2", etc
            as well as the recurrent hidden states under the key "rhs".
        """

        outputs = {}

        for i in range(self.num_critics_ext):
            value_net_name = "value_net{}".format(i + 1)
            value, rhs = self.get_value_specific_net(obs, rhs, done, value_net_name)
            outputs[value_net_name] = value

        for i in range(self.num_critics_int):
            value_net_name = "ivalue_net{}".format(i + 1)
            value, rhs = self.get_value_specific_net(obs, rhs, done, value_net_name)
            outputs[value_net_name] = value

        outputs["rhs"] = rhs
        return outputs

    def create_critic(self, name):
        """
        Create a critic value network and define it as class attribute under the name `name`.
        This actor defines defines value networks as:

        value = obs_feature_extractor + memory_net + v_prediction_layer

        and defines shared policy-value network as:
                                                     action_distribution
        value = obs_feature_extractor + memory_net +
                                                     v_prediction_layer

        Parameters
        ----------
        name : str
            Critic network name.
        """

        # If feature_extractor_network not defined, take default one based on input_space
        feature_extractor = self.feature_extractor_network or default_feature_extractor(self.input_space)

        if self.shared_policy_value_network:
            value_feature_extractor = nn.Identity()
            value_memory_net = nn.Identity()
            self.last_action_features = None

        else:

            # ---- 1. Define obs feature extractor ----------------------------

            value_feature_extractor = feature_extractor(
                self.input_space, **self.feature_extractor_kwargs)

            # ---- 2. Define memory network  ----------------------------------

            if isinstance(self.input_space, gym.spaces.Dict):
                dummy_obs = {k: torch.zeros(1, *self.input_space[k].shape) for k in self.input_space}
                features = value_feature_extractor(dummy_obs)
            else:
                dummy_obs = torch.zeros(1, *self.input_space.shape)
                features = value_feature_extractor(dummy_obs)

            if isinstance(value_feature_extractor, nn.Identity):
                feature_size = dummy_obs
            else:
                feature_size = features.shape[-1]

            if self.recurrent_net:
                value_memory_net = self.recurrent_net(feature_size, **self.recurrent_net_kwargs)
            else:
                value_memory_net = nn.Identity()

        # ---- 3. Define value predictor --------------------------------------

        init_ = lambda m: init(m, nn.init.orthogonal_, lambda x: nn.init.constant_(x, 0), gain=np.sqrt(0.01))
        value_predictor = init_(nn.Linear(self.recurrent_size, 1))

        # ---- 4. Concatenate all value net modules ---------------------------

        v_net = nn.Sequential(OrderedDict([
            ('feature_extractor', value_feature_extractor),
            ('memory_net', value_memory_net),
            ("predictor", value_predictor),
        ]))

        setattr(self, name, v_net)

    def create_policy(self, name):
        """
        Create a policy network and define it as class attribute under the name `name`.
        This actor defines policy network as:

        policy = obs_feature_extractor + memory_net + action_distribution

        Parameters
        ----------
        name : str
            Policy network name.
        """

        # If feature_extractor_network not defined, take default one based on input_space
        feature_extractor = self.feature_extractor_network or default_feature_extractor(self.input_space)

        # ---- 1. Define obs feature extractor --------------------------------

        policy_feature_extractor = feature_extractor(
            self.input_space, **self.feature_extractor_kwargs)

        # ---- 2. Define memory network  --------------------------------------

        if isinstance(self.input_space, gym.spaces.Dict):
            dummy_obs = {k: torch.zeros(1, *self.input_space[k].shape) for k in self.input_space}
            features = policy_feature_extractor(dummy_obs)
        else:
            dummy_obs = torch.zeros(1, *self.input_space.shape)
            features = policy_feature_extractor(dummy_obs)

        if isinstance(policy_feature_extractor, nn.Identity):
            feature_size = dummy_obs
        else:
            feature_size = features.shape[-1]

        if self.recurrent_net:
            policy_memory_net = self.recurrent_net(feature_size, **self.recurrent_net_kwargs)
            self.recurrent_size = policy_memory_net.recurrent_hidden_state_size
        else:
            policy_memory_net = nn.Identity()
            self.recurrent_size = feature_size

        # ---- 3. Define action distribution ----------------------------------

        if isinstance(self.action_space, gym.spaces.Discrete):
            dist = get_dist("Categorical")(self.recurrent_size, self.action_space.n)
            self.scale = None
            self.unscale = None

        elif isinstance(self.action_space, gym.spaces.Box):  # Continuous action space
            dist = get_dist("Gaussian")(self.recurrent_size, self.action_space.shape[0])
            self.scale = Scale(self.action_space)
            self.unscale = Unscale(self.action_space)

        elif isinstance(self.action_space, gym.spaces.Dict):
            raise NotImplementedError

        else:
            raise ValueError("Unrecognized action space")

        # ---- 4. Concatenate all policy modules ------------------------------

        policy_net = nn.Sequential(OrderedDict([
            ('feature_extractor', policy_feature_extractor),
            ('memory_net', policy_memory_net),
            ('dist', dist),
        ]))

        setattr(self, name, policy_net)



from minigrid.core.grid import Grid
from minigrid.core.mission import MissionSpace
from minigrid.core.world_object import Goal, Lava
from minigrid.minigrid_env import MiniGridEnv, WorldObj, COLORS, fill_coords, point_in_rect


class BlueGoal(WorldObj):
    def __init__(self):
        super().__init__("goal", "blue")
        self.goal_type = "blue_goal"

    def can_overlap(self):
        return True

    def render(self, img):
        fill_coords(img, point_in_rect(0, 1, 0, 1), COLORS[self.color])


class YellowGoal(WorldObj):
    def __init__(self):
        super().__init__("goal", "yellow")
        self.goal_type = "yellow_goal"

    def can_overlap(self):
        return True

    def render(self, img):
        fill_coords(img, point_in_rect(0, 1, 0, 1), COLORS[self.color])


class GreyGoal(WorldObj):
    def __init__(self):
        super().__init__("goal", "grey")
        self.goal_type = "grey_goal"

    def can_overlap(self):
        return True

    def render(self, img):
        fill_coords(img, point_in_rect(0, 1, 0, 1), COLORS[self.color])


class GreenGoal(WorldObj):
    def __init__(self):
        super().__init__("goal", "green")
        self.goal_type = "green_goal"

    def can_overlap(self):
        return True

    def render(self, img):
        fill_coords(img, point_in_rect(0, 1, 0, 1), COLORS[self.color])


class MultipleDeceivingRewardsEnv(MiniGridEnv):
    """
    ### Description

    Environment with 2 goals, one easy to get and one requiring more exploration.
    The agent must find the access to a room to get the second reward.

    ### Mission Space

    "get to a green goal square"

    ### Action Space

    | Num | Name         | Action       |
    |-----|--------------|--------------|
    | 0   | left         | Turn left    |
    | 1   | right        | Turn right   |
    | 2   | forward      | Move forward |
    | 3   | pickup       | Unused       |
    | 4   | drop         | Unused       |
    | 5   | toggle       | Unused       |
    | 6   | done         | Unused       |

    ### Observation Encoding

    - Each tile is encoded as a 3 dimensional tuple:
        `(OBJECT_IDX, COLOR_IDX, STATE)`
    - `OBJECT_TO_IDX` and `COLOR_TO_IDX` mapping can be found in
        [minigrid/minigrid.py](minigrid/minigrid.py)
    - `STATE` refers to the door state with 0=open, 1=closed and 2=locked

    ### Rewards

    A reward of '1' is given for the easy goal, and '0.5' for the hard goal.

    ### Termination

    The episode ends if any one of the following conditions is met:

    1. The agent reaches the goal.
    2. Timeout (see `max_steps`).

    ### Registered Configurations

    - `MiniGrid-DeceivingRewards-v0`
    """

    def __init__(self, **kwargs):

        mission_space = MissionSpace(
            mission_func=lambda: "get to the green goal square"
        )

        size = 40
        super().__init__(
            mission_space=mission_space,
            grid_size=size,
            max_steps=4 * size * size,
            # Set this to True for maximum speed
            see_through_walls=True,
            **kwargs
        )

    # generate a bunch of rooms

    def _gen_grid(self, width, height):

        # Create an empty grid
        self.grid = Grid(width, height)

        # Generate the surrounding walls
        self.grid.wall_rect(0, 0, width, height)

        # Create a room function
        def create_room(upper_right_corner, room_wall_length=8, door_pos=None):

            x, y = upper_right_corner

            self.grid.vert_wall(x, y, room_wall_length)
            self.grid.horz_wall(x, y, room_wall_length)
            self.grid.vert_wall(x + room_wall_length, y, room_wall_length)
            self.grid.horz_wall(x, y + room_wall_length, room_wall_length + 1)

            if door_pos:
                self.grid.set(*door_pos, None)

        # Create rooms
        create_room((4, 4), 8, (4, 8))
        create_room((20, 4), 8, (20, 8))
        create_room((20, 20), 8, (20, 24))

        # Create goals
        # Place a green goal square in the upper-right corner
        self.put_obj(GreenGoal(), width - 2, 1)

        # Place a blue goal square in the bottom-right corner
        self.put_obj(BlueGoal(), 8, 8)
        self.put_obj(YellowGoal(), 24, 8)
        self.put_obj(GreyGoal(), 24, 24)

        # Place the agent
        self.agent_pos = (width - 5, 7)
        self.agent_dir = 3

        self.mission = "get to the green goal square"

    def step(self, action):
        self.step_count += 1

        reward = 0
        terminated = False
        truncated = False

        # Get the position in front of the agent
        fwd_pos = self.front_pos

        # Get the contents of the cell in front of the agent
        fwd_cell = self.grid.get(*fwd_pos)

        # Rotate left
        if action == self.actions.left:
            self.agent_dir -= 1
            if self.agent_dir < 0:
                self.agent_dir += 4

        # Rotate right
        elif action == self.actions.right:
            self.agent_dir = (self.agent_dir + 1) % 4

        # Move forward
        elif action == self.actions.forward:
            if fwd_cell is None or fwd_cell.can_overlap():
                self.agent_pos = tuple(fwd_pos)
            if fwd_cell is not None and fwd_cell.type == "goal" and fwd_cell.goal_type == "yellow_goal":
                terminated = True
                reward = self._reward()

            if fwd_cell is not None and fwd_cell.type == "lava":
                terminated = True

        # Pick up an object
        elif action == self.actions.pickup:
            if fwd_cell and fwd_cell.can_pickup():
                if self.carrying is None:
                    self.carrying = fwd_cell
                    self.carrying.cur_pos = np.array([-1, -1])
                    self.grid.set(fwd_pos[0], fwd_pos[1], None)

        # Drop an object
        elif action == self.actions.drop:
            if not fwd_cell and self.carrying:
                self.grid.set(fwd_pos[0], fwd_pos[1], self.carrying)
                self.carrying.cur_pos = fwd_pos
                self.carrying = None

        # Toggle/activate an object
        elif action == self.actions.toggle:
            if fwd_cell:
                fwd_cell.toggle(self, fwd_pos)

        # Done action (not used by default)
        elif action == self.actions.done:
            pass

        else:
            raise ValueError(f"Unknown action: {action}")

        if self.step_count >= self.max_steps:
            truncated = True

        if self.render_mode == "human":
            self.render()

        obs = self.gen_obs()

        return obs, reward, terminated, truncated, {}

    def _reward(self):
        """
        Compute the reward to be given upon success
        """
        return 1



import os
import glob
import copy
import uuid
import torch
import numpy as np
from collections import defaultdict

import pytorchrl as prl
from pytorchrl.agent.storages.on_policy.gae_buffer import GAEBuffer as B


class PPODBuffer(B):
    """
    Storage class for PPO+D algorithm.

    Parameters
    ----------
    size : int
        Storage capacity along time axis.
    device: torch.device
        CPU or specific GPU where data tensors will be placed and class
        computations will take place. Should be the same device where the
        actor model is located.
    envs : VecEnv
        Vector of environments instance.
    actor : Actor
        Actor class instance.
    algorithm : Algorithm
        Algorithm class instance.
    initial_human_demos_dir : str
        Path to directory containing human initial demonstrations.
    initial_agent_demos_dir : str
        Path to directory containing other agent initial demonstrations.
    supplementary_demos_dir : str
        Path to a directory where additional demos can be added after training has started.
        these demos will be incorporated into the buffer as bonus agent demos.
    target_agent_demos_dir : str
        Path to directory where best reward demonstrations should be saved.
    rho : float
        PPO+D rho parameter.
    phi : float
        PPO+D phi parameter.
    alpha : float
        PPO+D alpha parameter
    gae_lambda : float
        GAE lambda parameter.
    total_buffer_demo_capacity : int
        Maximum number of demos to keep between reward and value demos.
    save_demos_every : int
        Save top demos every  `save_demo_frequency`th data collection.
    num_agent_demos_to_save : int
        Number of top reward demos to save.
    initial_reward_threshold : float
        initial value to use as reward threshold for new demos.
    demo_dtypes : dict
        data types to use for the demos.
    """

    # Accepted data fields. Inserting other fields will raise AssertionError
    on_policy_data_fields = prl.OnPolicyDataKeys

    # Data tensors to collect for each demos
    demos_data_fields = prl.DemosDataKeys

    def __init__(self, size, device, actor, algorithm, envs, rho=0.1, phi=0.3, gae_lambda=0.95, alpha=10,
                 total_buffer_demo_capacity=51, initial_human_demos_dir=None, initial_agent_demos_dir=None,
                 supplementary_demos_dir=None, target_agent_demos_dir=None, num_agent_demos_to_save=10,
                 initial_reward_threshold=None, save_demos_every=10,
                 demo_dtypes={prl.OBS: np.float32, prl.ACT: np.float32,  prl.REW: np.float32}):

        super(PPODBuffer, self).__init__(
            size=size,
            envs=envs,
            actor=actor,
            device=device,
            algorithm=algorithm,
            gae_lambda=gae_lambda,
        )

        # PPO + D parameters
        self.rho = rho
        self.phi = phi
        self.iter = 0
        self.alpha = alpha
        self.initial_rho = rho
        self.initial_phi = phi
        self.save_demos_every = save_demos_every
        self.max_demos = total_buffer_demo_capacity
        self.num_agent_demos_to_save = num_agent_demos_to_save
        self.initial_human_demos_dir = initial_human_demos_dir
        self.initial_agent_demos_dir = initial_agent_demos_dir
        self.supplementary_demos_dir = supplementary_demos_dir
        self.target_agent_demos_dir = target_agent_demos_dir

        # Data parameters
        self.demo_dtypes = demo_dtypes
        self.num_channels_obs = None  # Lazy initialization
        self.inserted_samples = 0

        # Reward and Value buffers
        self.reward_demos = []
        self.value_demos = []

        # Load initial demos
        self.load_initial_demos()

        # Define reward_threshold
        self.reward_threshold = initial_reward_threshold or - np.inf
        if len(self.reward_demos) > 0:
            new_threshold = min([d["TotalReward"] for d in self.reward_demos])
            if new_threshold > self.reward_threshold:
                self.reward_threshold = new_threshold
        self.max_demo_reward = max(
            [d["TotalReward"] for d in self.reward_demos]) if len(self.reward_demos) > 0 else -np.inf

        # Define variables to track potential demos
        self.potential_demos_val = {"env{}".format(i + 1): - np.inf for i in range(self.num_envs)}
        self.potential_demos = {"env{}".format(i + 1): defaultdict(list) for i in range(self.num_envs)}

        # Define variable to track demos in progress
        self.demos_in_progress = {
            "env{}".format(i + 1): {
                "ID": None,
                "Demo": None,
                "Step": 0,
                "DemoLength": -1,
                "MaxValue": - np.inf,
                prl.RHS: None,
            } for i in range(self.num_envs)}

        # To keep track of supplementary demos loaded
        self.supplementary_demos_loaded = []

    @classmethod
    def create_factory(cls, size, rho=0.1, phi=0.3, gae_lambda=0.95, alpha=10, total_buffer_demo_capacity=51,
                       initial_human_demos_dir=None, initial_agent_demos_dir=None, supplementary_demos_dir=None,
                       target_agent_demos_dir=None, num_agent_demos_to_save=10, initial_reward_threshold=None,
                       save_demos_every=10, demo_dtypes={prl.OBS: np.float32, prl.ACT: np.float32,  prl.REW: np.float32}):
        """
        Returns a function that creates PPODBuffer instances.

        Parameters
        ----------
        size : int
            Storage capacity along time axis.
        initial_human_demos_dir : str
            Path to directory containing human initial demonstrations.
        initial_agent_demos_dir : str
            Path to directory containing other agent initial demonstrations.
        supplementary_demos_dir : str
            Path to a directory where additional demos can be added after training has started.
            these demos will be incorporated into the buffer as bonus agent demos.
        target_agent_demos_dir : str
            Path to directory where best reward demonstrations should be saved.
        rho : float
            PPO+D rho parameter.
        phi : float
            PPO+D phi parameter.
        alpha : float
            PPO+D alpha parameter
        gae_lambda : float
            GAE lambda parameter.
        total_buffer_demo_capacity : int
            Maximum number of demos to keep between reward and value demos.
        save_demos_every : int
            Save top demos every  `save_demo_frequency`th data collection.
        num_agent_demos_to_save : int
            Number of top reward demos to save.
        initial_reward_threshold : float
            initial value to use as reward threshold for new demos.
        demo_dtypes : dict
            data types to use for the demos.

        Returns
        -------
        create_buffer_instance : func
            creates a new PPODBuffer class instance.
        """

        def create_buffer_instance(device, actor, algorithm, envs):
            """Create and return a PPODBuffer instance."""
            return cls(size, device, actor, algorithm, envs, rho, phi, gae_lambda, alpha, total_buffer_demo_capacity,
                       initial_human_demos_dir, initial_agent_demos_dir, supplementary_demos_dir, target_agent_demos_dir,
                       num_agent_demos_to_save, initial_reward_threshold, save_demos_every, demo_dtypes)
        return create_buffer_instance

    def before_gradients(self):
        """
        Before updating actor policy model, compute returns and advantages.
        """

        print("\nREWARD DEMOS {}, VALUE DEMOS {}, RHO {}, PHI {}, REWARD THRESHOLD {}, MAX DEMO REWARD {}\n".format(
            len(self.reward_demos), len(self.value_demos), self.rho, self.phi, self.reward_threshold, self.max_demo_reward))

        super(PPODBuffer, self).before_gradients()
        
        self.iter += 1
        if self.iter % self.save_demos_every == 0:
            self.save_demos()

        if self.supplementary_demos_dir:
            self.load_supplementary_demos()

    def after_gradients(self, batch, info):
        """
        After updating actor policy model, make sure self.step is at 0.
        Parameters
        ----------
        batch : dict
            Data batch used to compute the gradients.
        info : dict
            Additional relevant info from gradient computation.
        Returns
        -------
        info : dict
            info dict updated with relevant info from Storage.
        """

        super(PPODBuffer, self).after_gradients(batch, info)

        # info['NumberSamples'] -= self.inserted_samples
        self.inserted_samples = 0

        return info

    def insert_transition(self, sample):
        """
        Store new transition sample.

        Parameters
        ----------
        sample : dict
            Data sample (containing all tensors of an environment transition)
        """

        # Data tensors lazy initialization, only executed the first time
        if self.size == 0 and self.data[prl.OBS] is None:
            self.init_tensors(sample)
            self.get_num_channels_obs(sample)

        # Insert sample data
        for k in sample:

            if k not in self.storage_tensors:
                continue

            if not self.recurrent_actor and k in (prl.RHS, prl.RHS2):
                continue

            # We use the same tensor to store obs and obs2
            # We also use single tensors for rhs and rhs2,
            # and done and done2
            if k in (prl.OBS, prl.RHS, prl.DONE):
                pos = self.step + 1
                sample_k = "Next" + k
            else:
                pos = self.step
                sample_k = k

            # Copy sample tensor to buffer target position
            if isinstance(sample[k], dict):
                for x, v in sample[k].items():
                    self.data[k][x][pos].copy_(sample[sample_k][x])
            else:
                self.data[k][pos].copy_(sample[sample_k])

        # Track episodes for potential demos
        self.track_potential_demos(sample)

        all_envs, all_obs, all_done, all_rhs = [], [], [], {k: [] for k in sample[prl.RHS].keys()}

        for i in range(self.num_envs):

            if self.demos_in_progress["env{}".format(i + 1)]["Demo"]:

                # Get demo obs, rhs and done tensors to run forward pass
                obs = self.data[prl.OBS][self.step][i].unsqueeze(0)
                if self.demos_in_progress["env{}".format(i + 1)][prl.RHS]:
                    rhs = self.demos_in_progress["env{}".format(i + 1)][prl.RHS]
                    done = torch.zeros(1, 1).to(self.device)
                else:
                    obs, rhs, done = self.actor.actor_initial_states(obs)

                all_envs.append(i)
                all_obs.append(obs)
                all_done.append(done)
                for k in rhs:
                    all_rhs[k].append(rhs[k])

            # Otherwise check if end of episode reached and randomly start new demo
            elif sample[prl.DONE2][i] == 1.0:

                self.sample_demo(env_id=i)

        if len(all_envs) > 0:

            # Cast obs, rhs, dones into the right format
            all_obs = torch.cat(all_obs)
            all_done = torch.cat(all_done)
            for k in all_rhs:
                all_rhs[k] = torch.cat(all_rhs[k])

            # Run forward pass
            _, _, rhs2, algo_data = self.algo.acting_step(all_obs, all_rhs, all_done)

            for num, i in enumerate(all_envs):

                demo_step = self.demos_in_progress["env{}".format(i + 1)]["Step"]

                # Insert demo act and rew tensors to self.step
                for tensor in (prl.ACT, prl.REW):
                    self.data[tensor][self.step][i].copy_(
                        torch.FloatTensor(self.demos_in_progress["env{}".format(i + 1)]["Demo"][tensor][demo_step]))

                # Insert demo logprob to self.step. Demo action prob is 1.0, so logprob is 0.0
                self.data[prl.LOGP][self.step][i].copy_(torch.zeros(1))

                # Insert other tensors predicted in the forward pass
                for tensor in (prl.IREW, prl.VAL, prl.IVAL):
                    if tensor in algo_data.keys():
                        self.data[tensor][self.step][i].copy_(algo_data[tensor][num])

                # Update demo_in_progress variables
                self.demos_in_progress["env{}".format(i + 1)]["Step"] += 1

                self.demos_in_progress["env{}".format(i + 1)][prl.RHS] = {
                    k: rhs2[k][num].reshape(1, -1) for k in rhs2.keys()}

                self.demos_in_progress["env{}".format(i + 1)]["MaxValue"] = max(
                    [algo_data[prl.VAL][num].item(), self.demos_in_progress["env{}".format(i + 1)]["MaxValue"]])

                self.inserted_samples += 1

                # Handle end of demos
                if demo_step == self.demos_in_progress["env{}".format(i + 1)]["DemoLength"] - 1:

                    # If value demo
                    if "MaxValue" in self.demos_in_progress["env{}".format(i + 1)]["Demo"].keys():
                        for value_demo in self.value_demos:
                            # If demo still in buffer, update MaxValue
                            if self.demos_in_progress["env{}".format(i + 1)]["Demo"]["ID"] == value_demo["ID"]:
                                value_demo["MaxValue"] = self.demos_in_progress["env{}".format(i + 1)]["MaxValue"]

                    # Randomly sample new demos if last demos has finished
                    self.sample_demo(env_id=i)

                else:

                    # Insert demo done2 tensor to self.step + 1
                    self.data[prl.DONE][self.step + 1][i].copy_(torch.zeros(1))

                    # Insert demo obs2 tensor to self.step + 1
                    obs2 = torch.roll(all_obs[num:num+1], -self.num_channels_obs, dims=1).squeeze(0)
                    obs2[-self.num_channels_obs:].copy_(torch.FloatTensor(
                        self.demos_in_progress["env{}".format(i + 1)]["Demo"][prl.OBS][demo_step + 1]))
                    self.data[prl.OBS][self.step + 1][i].copy_(obs2)

                    # Insert demo rhs2 tensor to self.step + 1
                    if self.recurrent_actor:
                        for k in self.data[prl.RHS]:
                            self.data[prl.RHS][k][self.step + 1][i].copy_(rhs2[k][num].squeeze())

        self.step = (self.step + 1) % self.max_size
        self.size = min(self.size + 1, self.max_size)

    def track_potential_demos(self, sample):
        """ Tracks current episodes looking for potential demos """

        for i in range(self.num_envs):

            # Copy transition
            for tensor in self.demos_data_fields:
                if tensor in (prl.OBS):
                    self.potential_demos["env{}".format(i + 1)][tensor].append(copy.deepcopy(
                        sample[tensor][i, -self.num_channels_obs:]).cpu().numpy().astype(self.demo_dtypes[tensor]))
                else:
                    self.potential_demos["env{}".format(i + 1)][tensor].append(
                        copy.deepcopy(sample[tensor][i]).cpu().numpy().astype(self.demo_dtypes[tensor]))

            # Track highest value prediction
            self.potential_demos_val[i] = max([self.potential_demos_val["env{}".format(
                i + 1)], sample[prl.VAL][i].item()])

            # Handle end of episode
            if sample[prl.DONE2][i] == 1.0:

                # Get candidate demos
                potential_demo = {}
                for tensor in self.demos_data_fields:
                    potential_demo[tensor] = np.stack(self.potential_demos["env{}".format(i + 1)][tensor])
                    if tensor == prl.REW:
                        nonzero = np.flatnonzero(potential_demo[tensor] > 0.0)
                        last_reward = len(potential_demo[tensor]) if len(nonzero) == 0 else np.max(nonzero)

                # Cut off data after last reward
                for tensor in self.demos_data_fields:
                    potential_demo[tensor] = potential_demo[tensor][0:last_reward + 1]

                # Compute accumulated reward
                episode_reward = potential_demo[prl.REW].sum()
                potential_demo["ID"] = str(uuid.uuid4())
                potential_demo["TotalReward"] = episode_reward
                potential_demo["DemoLength"] = potential_demo[prl.ACT].shape[0]

                # Consider candidate demos for demos reward
                if episode_reward >= self.reward_threshold:

                    # If better than any other existing demo, empty buffer
                    if episode_reward > self.max_demo_reward:
                        self.reward_demos = []

                    # Add demos to reward buffer
                    self.reward_demos.append(potential_demo)

                    # Check if buffers are full
                    self.check_demo_buffer_capacity()

                    # Anneal rho and phi
                    self.anneal_parameters()

                    # Update reward_threshold.
                    self.reward_threshold = min([d["TotalReward"] for d in self.reward_demos])

                    # Update max demo reward
                    self.max_demo_reward = max([d["TotalReward"] for d in self.reward_demos])

                else:  # Consider candidate demos for value reward

                    # Find current number of demos, and current value threshold
                    potential_demo["MaxValue"] = self.potential_demos_val[i]
                    total_demos = len(self.reward_demos) + len(self.value_demos)
                    value_thresh = - np.float("Inf") if len(self.value_demos) == 0 \
                        else min([p["MaxValue"] for p in self.value_demos])

                    if self.potential_demos_val["env{}".format(i + 1)] >= value_thresh or total_demos < self.max_demos:

                        # Add demos to value buffer
                        self.value_demos.append(potential_demo)

                        # Check if buffers are full
                        self.check_demo_buffer_capacity()

                # Reset potential demos dict
                for tensor in self.demos_data_fields:
                    self.potential_demos["env{}".format(i + 1)][tensor] = []
                    self.potential_demos_val["env{}".format(i + 1)] = - np.inf

    def load_demo(self, demo_path):
        """Loads and returns a environment demonstration."""

        # Load demos tensors
        demo = np.load(demo_path)
        new_demo = {k: {} for k in self.demos_data_fields}

        if int(demo["FrameSkip"]) != self.frame_skip:
            raise ValueError(
                "Env and demo with different frame skip!")

        # Add action, obs, rew
        new_demo[prl.ACT] = demo[prl.ACT]
        new_demo[prl.OBS] = demo[prl.OBS]
        new_demo[prl.REW] = demo[prl.REW]

        new_demo.update({
            "ID": str(uuid.uuid4()),
            "DemoLength": demo[prl.ACT].shape[0],
            "TotalReward": new_demo[prl.REW].sum()})

        return new_demo

    def load_initial_demos(self):
        """
        Load initial demonstrations.
        Warning: make sure the environment frame_skip and frame_stack hyperparameters are
        the same as those used to record the demonstrations!
        """

        num_loaded_human_demos = 0
        num_loaded_reward_demos = 0

        initial_human_demos = glob.glob(self.initial_human_demos_dir + '/*.npz') if self.initial_human_demos_dir else []
        initial_reward_demos = glob.glob(self.initial_agent_demos_dir + '/*.npz') if self.initial_agent_demos_dir else []

        if len(initial_human_demos) + len(initial_reward_demos) > self.max_demos:
            raise ValueError("demo dir contains more than ´total_buffer_demo_capacity´")

        for demo_file in initial_human_demos:

            try:

                new_demo = self.load_demo(demo_file)
                self.reward_demos.append(new_demo)
                num_loaded_human_demos += 1

            except Exception:
                print("Failed to load human demo!")

        for demo_file in initial_reward_demos:

            try:

                new_demo = self.load_demo(demo_file)
                self.reward_demos.append(new_demo)
                num_loaded_reward_demos += 1

            except Exception:
                print("Failed to load agent demo!")

        self.num_loaded_human_demos = num_loaded_human_demos
        self.num_loaded_reward_demos = num_loaded_reward_demos
        print("\nLOADED {} HUMAN DEMOS AND {} REWARD DEMOS".format(num_loaded_human_demos, num_loaded_reward_demos))

    def load_supplementary_demos(self):
        """
        Load demonstrations found in the self.supplementary_demos (if any).
        Warning: make sure the environment frame_skip and frame_stack hyperparameters are
        the same as those used in the demonstrations!
        """

        # Create supp dir if necessary
        if not os.path.exists(self.supplementary_demos_dir):
            os.makedirs(self.supplementary_demos_dir, exist_ok=True)

        num_loaded_supplementary_demos = 0
        supplementary_demos = glob.glob(self.supplementary_demos_dir + '/*.npz') if self.supplementary_demos_dir else []

        for demo_file in supplementary_demos:

            if demo_file not in self.supplementary_demos_loaded:

                self.supplementary_demos_loaded.append(demo_file)
                try:

                    new_demo = self.load_demo(demo_file)
                    if new_demo["TotalReward"] >= self.reward_threshold:
                        self.reward_demos.append(new_demo)
                        num_loaded_supplementary_demos += 1

                    # Check if buffer is full and if so, handle it
                    self.check_demo_buffer_capacity()

                except Exception:
                    print("Failed to load supplementary demo!")

        print("\nLOADED {} SUPPLEMENTARY DEMOS\n".format(num_loaded_supplementary_demos))

    def sample_demo(self, env_id):
        """With probability rho insert reward demos, with probability phi insert value demos."""

        # Reset demos tracking variables
        self.demos_in_progress["env{}".format(env_id + 1)]["Step"] = 0
        self.demos_in_progress["env{}".format(env_id + 1)][prl.RHS] = None

        # Sample episode type
        episode_source = np.random.choice(["reward_demo", "value_demo", "env"],
            p=[self.rho, self.phi, 1.0 - self.rho - self.phi])

        if episode_source == "reward_demo" and len(self.reward_demos) > 0:

            # Randomly select a reward demo
            selected = np.random.choice(range(len(self.reward_demos)))
            demo = copy.deepcopy(self.reward_demos[selected])

        elif episode_source == "value_demo" and len(self.value_demos) > 0:

            # randomly select a value demo
            probs = np.array([p["MaxValue"] for p in self.value_demos]) ** self.alpha
            probs = probs / probs.sum()
            selected = np.random.choice(range(len(self.value_demos)), p=probs)
            demo = copy.deepcopy(self.value_demos[selected])

        else:
            demo = None

        # Set demos to demos_in_progress
        self.demos_in_progress["env{}".format(env_id + 1)]["Demo"] = demo

        # Set done to True
        self.data[prl.DONE][self.step + 1][env_id].copy_(torch.ones(1).to(self.device))

        # Set initial rhs to zeros
        if self.recurrent_actor:
            for k in self.data[prl.RHS]:
                self.data[prl.RHS][k][self.step + 1][env_id].fill_(0.0)

        if demo:

            # Set demos length
            self.demos_in_progress["env{}".format(env_id + 1)]["DemoLength"] = demo["DemoLength"]

            # Set demos MaxValue
            self.demos_in_progress["env{}".format(env_id + 1)]["MaxValue"] = - np.Inf

            # Set next buffer obs to be the starting demo obs
            for k in range(self.frame_stack):
                self.data[prl.OBS][self.step + 1][env_id][
                k * self.num_channels_obs:(k + 1) * self.num_channels_obs].copy_(torch.FloatTensor(
                    self.demos_in_progress["env{}".format(env_id + 1)]["Demo"][prl.OBS][0]))

        else:
            # Reset `i-th` environment as set next buffer obs to be the starting episode obs
            self.data[prl.OBS][self.step + 1][env_id].copy_(self.envs.reset_single_env(env_id=env_id).squeeze())

            # Reset potential demos dict
            for tensor in self.demos_data_fields:
                self.potential_demos["env{}".format(env_id + 1)][tensor] = []
                self.potential_demos_val["env{}".format(env_id + 1)] = - np.inf

    def anneal_parameters(self):
        """Update demos probabilities as explained in PPO+D paper."""

        if 0.0 < self.rho < 1.0 and len(self.value_demos) > 0:
            self.rho += self.initial_phi / len(self.value_demos)
            self.rho = np.clip(self.rho, 0.0, self.initial_rho + self.initial_phi)

        if 0.0 < self.phi < 1.0 and len(self.value_demos) > 0:
            self.phi -= self.initial_phi / len(self.value_demos)
            self.phi = np.clip(self.phi, 0.0, self.initial_rho + self.initial_phi)

    def check_demo_buffer_capacity(self):
        """
        Check total amount of demos. If total amount of demos exceeds
        self.max_demos, pop demos.
        """

        # First pop value demos
        total_demos = len(self.reward_demos) + len(self.value_demos)
        if total_demos > self.max_demos:
            for _ in range(min(total_demos - self.max_demos, len(self.value_demos))):

                # Pop value demos with lowest MaxValue
                del self.value_demos[np.array([p["MaxValue"] for p in self.value_demos]).argmin()]

        # If after popping all value demos, still over max_demos, pop reward demos
        if len(self.reward_demos) > self.max_demos:
            for _ in range(len(self.reward_demos) - self.max_demos):

                # # Option 1: Eject agent demo with lowest reward
                # rewards = np.array([p[prl.REW].sum() for p in self.reward_demos[self.num_loaded_human_demos:]])
                # del self.reward_demos[np.argmin(rewards) + self.num_loaded_human_demos]

                # Option 2: Eject longer agent demo
                lengths = np.array([p[prl.OBS].shape[0] for p in self.reward_demos[self.num_loaded_human_demos:]])
                del self.reward_demos[np.argmax(lengths) + self.num_loaded_human_demos]

    def save_demos(self):
        """
        Saves the top `num_rewards_demos` demos from the reward demos buffer and
        the top `num_value_demos` demos from the value demos buffer.
        """

        if self.target_agent_demos_dir:

            # Create target dir for reward demos if necessary
            if not os.path.exists(self.target_agent_demos_dir):
                os.makedirs(self.target_agent_demos_dir, exist_ok=True)

            # Rank agent demos according to episode reward
            reward_ranking = np.flip(np.array(
                [d["TotalReward"] for d in self.reward_demos[self.num_loaded_human_demos:]]
            ).argsort())[:self.num_agent_demos_to_save]

            # Save agent reward demos
            for num, demo_pos in enumerate(reward_ranking):
                filename = "reward_demo_{}".format(num + 1)
                demo_pos += self.num_loaded_human_demos
                np.savez(
                    os.path.join(self.target_agent_demos_dir, filename),
                    Observation=np.array(self.reward_demos[demo_pos][prl.OBS]).astype(self.demo_dtypes[prl.OBS]),
                    Reward=np.array(self.reward_demos[demo_pos][prl.REW]).astype(self.demo_dtypes[prl.REW]),
                    Action=np.array(self.reward_demos[demo_pos][prl.ACT]).astype(self.demo_dtypes[prl.ACT]),
                    FrameSkip=self.frame_skip)




import torch.nn as nn


def get_gain(activation):
    name = activation.__name__.lower()
    if name == "leakyrelu":
        name = "leaky_relu"
    gain = nn.init.calculate_gain(name)
    return gain



import numpy as np
from pytorchrl.envs.generative_chemistry.diversity_filter.base_diversity_filter import BaseDiversityFilter


class NoFilterWithPenalty(BaseDiversityFilter):
    """Penalize repeatedly generated compounds."""

    def __init__(self, minscore=0.4, bucket_size=25, minsimilarity=0.0):
        super().__init__(minscore, bucket_size, minsimilarity)

    def update_score(self, score, smile) -> np.array:

        if smile is not None:
            smile = self._chemistry.convert_to_rdkit_smiles(smile)
            score = 0.5 * score if self._smiles_exists(smile) else score

        if score >= self.minscore:
            self._add_to_memory(smile)

        return score



import random
import numpy as np
from copy import deepcopy

import torch
import torch.nn as nn
import torch.optim as optim
import torch.nn.functional as F

import pytorchrl as prl
from pytorchrl.agent.algorithms.base import Algorithm
from pytorchrl.agent.algorithms.utils import get_gradients, set_gradients
from pytorchrl.agent.algorithms.policy_loss_addons import PolicyLossAddOn


class DDQN(Algorithm):
    """
    Deep Q Learning algorithm class.

    Algorithm class to execute DQN, from Mhin et al.
    (https://www.nature.com/articles/nature14236?wm=book_wap_0005) with
    target network.

    Parameters
    ----------
    device : torch.device
        CPU or specific GPU where class computations will take place.
    envs : VecEnv
        Vector of environments instance.
    actor : ActorCritic
        actor class instance.
    lr : float
        learning rate.
    gamma : float
        Discount factor parameter.
    num_updates: int
        Num consecutive actor updates before data collection continues.
    update_every: int
        Regularity of actor updates in number environment steps.
    start_steps: int
        Num of initial random environment steps before learning starts.
    mini_batch_size: int
        Size of actor update batches.
    num_test_episodes : int
        Number of episodes to complete in each test phase.
    test_every : int
        Regularity of test evaluations in actor updates.
    max_grad_norm : float
        Gradient clipping parameter.
    initial_epsilon : float
        initial value for DQN epsilon parameter.
    epsilon_decay : float
        Exponential decay rate for epsilon parameter.
    policy_loss_addons : list
        List of PolicyLossAddOn components adding loss terms to the algorithm policy loss.
    """

    def __init__(self,
                 device,
                 envs,
                 actor,
                 lr=1e-4,
                 gamma=0.99,
                 polyak=0.995,
                 num_updates=1,
                 update_every=50,
                 test_every=5000,
                 max_grad_norm=0.5,
                 start_steps=20000,
                 mini_batch_size=64,
                 num_test_episodes=5,
                 initial_epsilon=1.0,
                 epsilon_decay=0.999,
                 target_update_interval=1,
                 policy_loss_addons=[]):

        # ---- General algo attributes ----------------------------------------

        # Discount factor
        self._gamma = gamma

        # Number of steps collected with initial random policy
        self._start_steps = int(start_steps)

        # Times data in the buffer is re-used before data collection proceeds
        self._num_epochs = 1  # Default to 1 for off-policy algorithms

        # Number of data samples collected between network update stages
        self._update_every = int(update_every)

        # Number mini batches per epoch
        self._num_mini_batch = int(num_updates)

        # Size of update mini batches
        self._mini_batch_size = int(mini_batch_size)

        # Number of network updates between test evaluations
        self._test_every = int(test_every)

        # Number of episodes to complete when testing
        self._num_test_episodes = int(num_test_episodes)

        # ---- DDQN-specific attributes ---------------------------------------

        self.iter = 0
        self.envs = envs
        self.actor = actor
        self.device = device
        self.polyak = polyak
        self.epsilon = initial_epsilon
        self.max_grad_norm = max_grad_norm
        self.epsilon_decay = epsilon_decay
        self.target_update_interval = target_update_interval

        assert hasattr(self.actor, "q1"), "DDPG requires q critic (num_critics=1)"

        # Create target network
        self.actor_targ = deepcopy(actor)

        # Freeze target networks with respect to optimizers
        for p in self.actor_targ.parameters():
            p.requires_grad = False

        # ----- Policy Loss Addons --------------------------------------------

        # Sanity check, policy_loss_addons is a PolicyLossAddOn instance
        # or a list of PolicyLossAddOn instances
        assert isinstance(policy_loss_addons, (PolicyLossAddOn, list)),\
            "DDQN policy_loss_addons parameter should be a  PolicyLossAddOn instance " \
            "or a list of PolicyLossAddOn instances"
        if isinstance(policy_loss_addons, list):
            for addon in policy_loss_addons:
                assert isinstance(addon, PolicyLossAddOn), \
                    "DDQN policy_loss_addons parameter should be a  PolicyLossAddOn" \
                    " instance or a list of PolicyLossAddOn instances"
        else:
            policy_loss_addons = [policy_loss_addons]

        self.policy_loss_addons = policy_loss_addons
        for addon in self.policy_loss_addons:
            addon.setup(self.actor, self.device)

        # ----- Optimizer -----------------------------------------------------

        self.q_optimizer = optim.Adam(self.actor.q1.parameters(), lr=lr)

    @classmethod
    def create_factory(cls,
                       lr=1e-4,
                       gamma=0.99,
                       polyak=0.995,
                       num_updates=50,
                       update_every=50,
                       test_every=5000,
                       start_steps=20000,
                       max_grad_norm=0.5,
                       mini_batch_size=64,
                       num_test_episodes=5,
                       epsilon_decay=0.999,
                       initial_epsilon=1.0,
                       target_update_interval=1,
                       policy_loss_addons=[]):
        """
        Returns a function to create new DDQN instances.

        Parameters
        ----------
        lr : float
            learning rate.
        gamma : float
            Discount factor parameter.
        polyak: float
            Polyak averaging parameter.
        num_updates: int
            Num consecutive actor updates before data collection continues.
        update_every: int
            Regularity of actor updates in number environment steps.
        start_steps: int
            Num of initial random environment steps before learning starts.
        mini_batch_size: int
            Size of actor update batches.
        target_update_interval: float
            regularity of target nets updates with respect to actor Adam updates.
        num_test_episodes : int
            Number of episodes to complete in each test phase.
        test_every : int
            Regularity of test evaluations in actor updates.
        max_grad_norm : float
            Gradient clipping parameter.
        initial_epsilon : float
            initial value for DQN epsilon parameter.
        epsilon_decay : float
            Exponential decay rate for epsilon parameter.
        policy_loss_addons : list
            List of PolicyLossAddOn components adding loss terms to the algorithm policy loss.

        Returns
        -------
        create_algo_instance : func
            Function that creates a new DDQN class instance.
        algo_name : str
            Name of the algorithm.
        """

        def create_algo_instance(device, actor, envs):
            return cls(lr=lr,
                       envs=envs,
                       actor=actor,
                       gamma=gamma,
                       device=device,
                       polyak=polyak,
                       test_every=test_every,
                       start_steps=start_steps,
                       num_updates=num_updates,
                       update_every=update_every,
                       epsilon_decay=epsilon_decay,
                       max_grad_norm=max_grad_norm,
                       mini_batch_size=mini_batch_size,
                       initial_epsilon=initial_epsilon,
                       num_test_episodes=num_test_episodes,
                       target_update_interval=target_update_interval,
                       policy_loss_addons=policy_loss_addons)

        return create_algo_instance, prl.DDQN

    def acting_step(self, obs, rhs, done, deterministic=False):
        """
        DDQN acting function.

        Parameters
        ----------
        obs: torch.tensor
            Current world observation
        rhs: dict
            RNN recurrent hidden states.
        done: torch.tensor
            1.0 if current obs is the last one in the episode, else 0.0.
        deterministic: bool
            Whether to randomly sample action from predicted distribution or taking the mode.

        Returns
        -------
        action: torch.tensor
            Predicted next action.
        clipped_action: torch.tensor
            Predicted next action (clipped to be within action space).
        rhs: batch
            Actor recurrent hidden state.
        other: dict
            Additional DDQN predictions, which are not used in other algorithms.
        """

        # Epsilon-greedy action selection
        if random.random() > self.epsilon:
            with torch.no_grad():
                q = self.actor.get_q_scores(obs).get("q1")
                action = clipped_action = torch.argmax(q, dim=1).unsqueeze(0)
        else:
            action = clipped_action = torch.tensor(
                [self.actor.action_space.sample()]).unsqueeze(0)

        other = {}
        return action, clipped_action, rhs, other

    def compute_loss(self, batch, n_step=1, per_weights=1):
        """
        Calculate DDQN loss

        Parameters
        ----------
        batch: dict
            Data batch dict containing all required tensors to compute DDQN loss.

        Returns
        -------
        loss : torch.tensor
            DDQN loss.
        errors : torch.tensor
            TD errors.
        """

        o, rhs, d = batch[prl.OBS], batch[prl.RHS], batch[prl.DONE]
        a, r = batch[prl.ACT], batch[prl.REW]
        o2, rhs2, d2 = batch[prl.OBS2], batch[prl.RHS2], batch[prl.DONE2]

        # Get max predicted Q values (for next states) from target model
        q_targ_vals = self.actor_targ.get_q_scores(o2, rhs2, d2).get("q1")
        q_targ_next = q_targ_vals.max(dim=1)[0].unsqueeze(1)

        # Compute Q targets for current states
        q_targ = r + (self.gamma ** n_step) * (1 - d2) * q_targ_next

        # Get expected Q values from local model
        q_vals = self.actor.get_q_scores(o, rhs, d).get("q1")
        q_exp = q_vals.gather(1, a.long())

        # Compute loss
        loss = F.mse_loss(q_targ, q_exp)

        errors = (q_exp - q_targ).abs().detach().cpu()

        # Extend policy loss with addons
        addons_info = {}
        for addon in self.policy_loss_addons:
            addon_loss, addons_info = addon.compute_loss_term(data, dist, addons_info)
            loss += addon_loss

        return loss, errors, addons_info

    def compute_gradients(self, batch, grads_to_cpu=True):
        """
        Compute loss and compute gradients but don't do optimization step,
        return gradients instead.

        Parameters
        ----------
        data: dict
            data batch containing all required tensors to compute DQN loss.
        grads_to_cpu: bool
            If gradient tensor will be sent to another node, need to be in CPU.

        Returns
        -------
        grads: list of tensors
            List of actor gradients.
        info: dict
            Dict containing current DQN iteration information.
        """

        # Recurrent burn-in
        if self.actor.is_recurrent:
            batch = self.actor.burn_in_recurrent_states(batch)

        # N-step returns
        n_step = batch["n_step"] if "n_step" in batch else 1.0

        # PER
        per_weights = batch["per_weights"] if "per_weights" in batch else 1.0

        # Compute DDQN loss and gradients
        loss, errors, addons_info = self.compute_loss(batch, n_step, per_weights)
        self.q_optimizer.zero_grad()
        loss.backward()
        nn.utils.clip_grad_norm_(self.actor.q1.parameters(), self.max_grad_norm)
        grads = get_gradients(self.actor.q1, grads_to_cpu=grads_to_cpu)

        info = {
            "loss_q": loss.detach().item(),
            "epsilon": self.epsilon,
        }

        if "per_weights" in batch:
            info.update({"errors": errors})

        info.update(addons_info)

        return grads, info

    def update_target_networks(self):
        """Update actor critic target networks with polyak averaging"""
        if self.iter % self.target_update_interval == 0:
            with torch.no_grad():
                for p, p_targ in zip(self.actor.parameters(), self.actor_targ.parameters()):
                    p_targ.data.mul_(self.polyak)
                    p_targ.data.add_((1 - self.polyak) * p.data)

    def update_epsilon(self):
        self.epsilon *= self.epsilon_decay
        self.epsilon = np.clip(self.epsilon, 0.05, 1.0)

    def apply_gradients(self, gradients=None):
        """
        Take an optimization step, previously setting new gradients if provided.
        Update also target networks.

        Parameters
        ----------
        gradients: list of tensors
            List of actor gradients.
        """
        if gradients is not None:
            set_gradients(
                self.actor.q1, gradients=gradients, device=self.device)
        self.q_optimizer.step()

        # Update target networks by polyak averaging.
        self.iter += 1
        self.update_target_networks()
        self.update_epsilon()

    def set_weights(self, weights):
        """
        Update actor critic with the given weights. Update also target networks.

        Parameters
        ----------
        weights: dict of tensors
            Dict containing actor weights to be set.
        """
        self.actor.load_state_dict(weights)

        # Update target networks by polyak averaging.
        self.iter += 1
        self.update_target_networks()
        self.update_epsilon()

    def update_algo_parameter(self, parameter_name, new_parameter_value):
        """
        If `parameter_name` is an attribute of the algorithm, change its value
        to `new_parameter_value value`.

        Parameters
        ----------
        parameter_name : str
            Worker.algo attribute name
        new_parameter_value : int or float
            New value for `parameter_name`.
        """
        if hasattr(self, parameter_name):
            setattr(self, parameter_name, new_parameter_value)
        if parameter_name == "lr":
            for param_group in self.q_optimizer.param_groups:
                param_group['lr'] = new_parameter_value



import os
import glob
import copy
import uuid
import torch
import numpy as np
from copy import deepcopy
from collections import defaultdict

import pytorchrl as prl
from pytorchrl.agent.storages.on_policy.gae_buffer import GAEBuffer as B

MIN_BUFFER_SIZE = 10


class PPOD2Buffer(B):
    """
    Storage class for a modified version of the PPO+D algorithm.

    This version of the PPO+D buffer uses intrinsic rewards (if available) instead of value predictions to rank
    potentially interesting demos. Additionally, this version does not differentiate between human demos and agent
    demos, only between reward demos (independently of their origin) and intrinsic demos. If no intrinsic rewards
    are available, only reward demos are replayed.

    Parameters
    ----------
    size : int
        Storage capacity along time axis.
    device: torch.device
        CPU or specific GPU where data tensors will be placed and class
        computations will take place. Should be the same device where the
        actor model is located.
    actor : Actor
        Actor class instance.
    algorithm : Algorithm
        Algorithm class instance.
    envs : VecEnv
        Vector of experiments instance.
    rho : float
        PPO+D rho parameter.
    phi : float
        PPO+D phi parameter.
    gae_lambda : float
        GAE lambda parameter.
    alpha : float
        PPO+D alpha parameter
    initial_reward_threshold : float
        initial intrinsic to use as reward threshold for new reward_demos.
    initial_reward_demos_dir : str
        Path to directory containing agent initial demonstrations.
    supplementary_demos_dir : str
        Path to a directory where additional demos can be added after training has started.
        these demos will be incorporated into the buffer as bonus agent demos.
    target_reward_demos_dir : str
        Path to directory where best reward demonstrations should be saved.
    num_reward_demos_to_save : int
        Number of top reward reward_demos to save.
    initial_int_demos_dir : str
        Path to directory containing intrinsic initial demonstrations.
    target_int_demos_dir : str
        Path to directory where best intrinsic demonstrations should be saved.
    num_int_demos_to_save : int
        Number of top intrinsic reward_demos to save.
    total_buffer_demo_capacity : int
        Maximum number of reward_demos to keep between reward and intrinsic reward_demos.
    save_demos_every : int
        Save top reward_demos every  `save_demo_frequency`th data collection.
    demo_dtypes : dict
        data types to use for the reward_demos.
    """

    # Accepted data fields. Inserting other fields will raise AssertionError
    storage_tensors = prl.OnPolicyDataKeys

    # Data tensors to collect for each reward_demos
    demos_data_fields = prl.DemosDataKeys

    def __init__(self, size, device, actor, algorithm, envs, rho=0.05, phi=0.05, gae_lambda=0.95, alpha=10,
                 total_buffer_demo_capacity=50, initial_reward_threshold=None, initial_reward_demos_dir=None,
                 supplementary_demos_dir=None, target_reward_demos_dir=None, num_reward_demos_to_save=None,
                 initial_int_demos_dir=None, target_int_demos_dir=None, num_int_demos_to_save=None, save_demos_every=10,
                 demo_dtypes={prl.OBS: np.float32, prl.ACT: np.float32,  prl.REW: np.float32}):

        super(PPOD2Buffer, self).__init__(
            size=size,
            envs=envs,
            actor=actor,
            device=device,
            algorithm=algorithm,
            gae_lambda=gae_lambda,
        )

        # PPO + D parameters
        self.rho = rho
        self.phi = phi
        self.alpha = alpha
        self.iter = 0
        self.eta = 0.9  # To weight max and cumsum intrinsic rewards

        # Define buffer demos capacity
        if rho != 0.0 or phi != 0.0:
            self.max_reward_demos = max(MIN_BUFFER_SIZE, int(total_buffer_demo_capacity * (rho / (rho + phi))))
            self.max_intrinsic_demos = max(MIN_BUFFER_SIZE, int(total_buffer_demo_capacity * (phi / (rho + phi))))
        else:
            self.max_reward_demos = MIN_BUFFER_SIZE
            self.max_intrinsic_demos = MIN_BUFFER_SIZE

        # Demo - related parameters
        self.save_demos_every = save_demos_every
        self.max_demos = total_buffer_demo_capacity
        self.num_reward_demos_to_save = num_reward_demos_to_save or self.max_reward_demos
        self.initial_reward_demos_dir = initial_reward_demos_dir
        self.target_reward_demos_dir = target_reward_demos_dir
        self.num_int_demos_to_save = num_int_demos_to_save or self.max_intrinsic_demos
        self.initial_int_demos_dir = initial_int_demos_dir
        self.supplementary_demos_dir = supplementary_demos_dir
        self.target_int_demos_dir = target_int_demos_dir

        # Data parameters
        self.demo_dtypes = demo_dtypes
        self.num_channels_obs = None  # Lazy initialization
        self.inserted_samples = 0
        self.storage_tensors += (prl.MASK, )

        # Define reward and intrinsic buffers
        self.reward_demos = []
        self.intrinsic_demos = []

        # Load initial reward_demos
        self.load_initial_demos()

        # Define reward_threshold
        self.reward_threshold = initial_reward_threshold or - np.inf
        if len(self.reward_demos) > 0:
            new_threshold = min([d["TotalReward"] for d in self.reward_demos])
            if new_threshold > self.reward_threshold:
                self.reward_threshold = new_threshold
        self.max_demo_reward = max(
            [d["TotalReward"] for d in self.reward_demos]) if len(self.reward_demos) > 0 else -np.inf

        # Define intrinsic_threshold
        self.intrinsic_threshold = - np.inf
        if len(self.intrinsic_demos) > 0:
            self.intrinsic_threshold = min([d["IntrinsicReward"] for d in self.intrinsic_demos])

        # Define variables to track potentially interesting demos
        self.potential_demos_cumsum_int = {"env{}".format(i + 1): 0.0 for i in range(self.num_envs)}
        self.potential_demos_max_int = {"env{}".format(i + 1): 0.0 for i in range(self.num_envs)}
        self.potential_demos = {"env{}".format(i + 1): defaultdict(list) for i in range(self.num_envs)}

        # Define variable to track reward_demos in progress
        self.demos_in_progress = {
            "env{}".format(i + 1): {
                "ID": None, "Demo": None, "Step": 0, "DemoLength": -1, "DemoType": None,
                "CumIntrinsicReward": 0.0, "MaxIntrinsicReward": 0.0, prl.RHS: None,
            } for i in range(self.num_envs)}

        # To keep track of supplementary demos loaded
        self.supplementary_demos_loaded = []

    @classmethod
    def create_factory(cls, size, rho=0.05, phi=0.05, gae_lambda=0.95, alpha=10, total_buffer_demo_capacity=50,
                       initial_reward_threshold=None, initial_reward_demos_dir=None,
                       supplementary_demos_dir=None, target_reward_demos_dir=None, num_reward_demos_to_save=None,
                       initial_int_demos_dir=None, target_int_demos_dir=None, num_int_demos_to_save=None,
                       save_demos_every=10, demo_dtypes={prl.OBS: np.float32, prl.ACT: np.float32, prl.REW: np.float32}):
        """
        Returns a function that creates PPOD2Buffer instances.

        Parameters
        ----------
        size : int
            Storage capacity along time axis.
        rho : float
            PPO+D rho parameter.
        phi : float
            PPO+D phi parameter.
        gae_lambda : float
            GAE lambda parameter.
        alpha : float
            PPO+D alpha parameter
        total_buffer_demo_capacity : int
            Maximum number of reward_demos to keep between reward and intrinsic reward_demos.
            initial_reward_threshold : float
        initial_reward_threshold : float
            initial intrinsic to use as reward threshold for new reward_demos.
        initial_reward_demos_dir : str
            Path to directory containing agent initial demonstrations.
        supplementary_demos_dir : str
            Path to a directory where additional demos can be added after training has started.
            these demos will be incorporated into the buffer as bonus agent demos.
        target_reward_demos_dir : str
            Path to directory where best reward demonstrations should be saved.
        num_reward_demos_to_save : int
            Number of top reward reward_demos to save.
        initial_int_demos_dir : str
            Path to directory containing intrinsic initial demonstrations.
        target_int_demos_dir : str
            Path to directory where best intrinsic demonstrations should be saved.
        num_int_demos_to_save : int
            Number of top intrinsic reward_demos to save.
        total_buffer_demo_capacity : int
            Maximum number of reward_demos to keep between reward and intrinsic reward_demos.
        save_demos_every : int
            Save top reward_demos every  `save_demo_frequency`th data collection.
        demo_dtypes : dict
            data types to use for the reward_demos.

        Returns
        -------
        create_buffer_instance : func
            creates a new PPOD2Buffer class instance.
        """
        def create_buffer_instance(device, actor, algorithm, envs):
            """Create and return a PPOD2Buffer instance."""
            return cls(size, device, actor, algorithm, envs, rho, phi, gae_lambda, alpha,
                       total_buffer_demo_capacity, initial_reward_threshold, initial_reward_demos_dir,
                       supplementary_demos_dir, target_reward_demos_dir, num_reward_demos_to_save,
                       initial_int_demos_dir, target_int_demos_dir, num_int_demos_to_save,
                       save_demos_every, demo_dtypes)
        return create_buffer_instance

    def init_tensors(self, sample):
        """
        Lazy initialization of data tensors from a sample.

        Parameters
        ----------
        sample : dict
            Data sample (containing all tensors of an environment transition)
        """

        if prl.EMBED in sample.keys():
            self.demos_data_fields += (prl.EMBED,)
            self.storage_tensors += (prl.EMBED,)
            self.demo_dtypes[prl.EMBED] = sample[prl.EMBED].cpu().numpy().dtype

        super(PPOD2Buffer, self).init_tensors(sample)

        if prl.IREW not in sample.keys():
            self.phi = 0.0
            self.max_reward_demos = self.max_demos
            self.max_intrinsic_demos = 0

    def before_gradients(self):
        """Before updating actor policy model, compute returns and advantages."""

        super(PPOD2Buffer, self).before_gradients()
        print("\nREWARD DEMOS {}, INTRINSIC DEMOS {}, RHO {}, PHI {}, REWARD THRESHOLD {}, MAX DEMO REWARD {},"
              " INTRINSIC THRESHOLD {}\n".format(len(self.reward_demos), len(self.intrinsic_demos),
            self.rho, self.phi, self.reward_threshold, self.max_demo_reward, self.intrinsic_threshold))

        self.iter += 1
        if self.iter % self.save_demos_every == 0:
            self.save_demos()

        if self.supplementary_demos_dir:
            self.load_supplementary_demos()

    def after_gradients(self, batch, info):
        """
        After updating actor policy model, make sure self.step is at 0.

        Parameters
        ----------
        batch : dict
            Data batch used to compute the gradients.
        info : dict
            Additional relevant info from gradient computation.

        Returns
        -------
        info : dict
            info dict updated with relevant info from Storage.
        """

        super(PPOD2Buffer, self).after_gradients(batch, info)

        # info['NumberSamples'] -= self.inserted_samples
        self.inserted_samples = 0

        return info

    def insert_transition(self, sample):
        """
        Store new transition sample.

        Parameters
        ----------
        sample : dict
            Data sample (containing all tensors of an environment transition)
        """

        # If obs embeddings available, add them (for logging purposes)
        if prl.EMBED in sample[prl.INFO][0].keys():
            obs_embeds = [i[prl.EMBED] for i in sample[prl.INFO]]
            obs_embeds = np.stack(obs_embeds).reshape(len(obs_embeds), -1)
            sample.update({prl.EMBED: torch.from_numpy(obs_embeds)})

        if self.num_channels_obs is None:
            self.get_num_channels_obs(sample)

        # Data tensors lazy initialization, only executed the first time
        if self.size == 0 and self.data[prl.OBS] is None:
            self.init_tensors(sample)

        # Track episodes for potential agent_demos
        self.track_potential_demos(sample)

        # Insert environment sample data
        for k in sample:

            if k not in self.storage_tensors:
                continue

            if not self.recurrent_actor and k in (prl.RHS, prl.RHS2):
                continue

            # We use the same tensor to store obs and obs2
            # We also use single tensors for rhs and rhs2,
            # and done and done2
            if k in (prl.OBS, prl.RHS, prl.DONE):
                pos = self.step + 1
                sample_k = "Next" + k
            else:
                pos = self.step
                sample_k = k

            # Copy sample tensor to buffer target position
            if isinstance(sample[k], dict):
                for x, v in sample[k].items():
                    self.data[k][x][pos].copy_(sample[sample_k][x])
            else:
                self.data[k][pos].copy_(sample[sample_k])

        # Set inserted data mask to 1.0
        if prl.MASK in self.data.keys():
            self.data[prl.MASK][self.step].fill_(1.0)

        # Overwrite demo data in environments where a demo being replayed
        all_envs, all_obs, all_done, all_rhs = [], [], [], {k: [] for k in sample[prl.RHS].keys()}

        # Step 1: prepare tensors
        for i in range(self.num_envs):

            # If demo replay is in progress
            if self.demos_in_progress["env{}".format(i + 1)]["Demo"]:

                # Get demo obs, rhs and done tensors to run forward pass
                obs = self.data[prl.OBS][self.step][i].unsqueeze(0)
                if self.demos_in_progress["env{}".format(i + 1)][prl.RHS]:
                    rhs = self.demos_in_progress["env{}".format(i + 1)][prl.RHS]
                    done = torch.zeros(1, 1).to(self.device)
                else:
                    obs, rhs, done = self.actor.actor_initial_states(obs)

                all_envs.append(i)
                all_obs.append(obs)
                all_done.append(done)
                for k in rhs:
                    all_rhs[k].append(rhs[k])

            # Otherwise check if end of episode reached and randomly start new demo
            elif sample[prl.DONE2][i] == 1.0:

                self.sample_demo(env_id=i)

        # Step 2: perform acting step and overwrite data
        if len(all_envs) > 0:

            # Cast obs, rhs, dones into the right format
            all_obs = torch.cat(all_obs)
            all_done = torch.cat(all_done)
            for k in all_rhs:
                all_rhs[k] = torch.cat(all_rhs[k])

            # Run forward pass
            _, _, rhs2, algo_data = self.algo.acting_step(all_obs, all_rhs, all_done)

            for num, i in enumerate(all_envs):

                demo_step = self.demos_in_progress["env{}".format(i + 1)]["Step"]

                # Insert demo act and rew tensors to self.step
                for tensor in (prl.ACT, prl.REW):
                    self.data[tensor][self.step][i].copy_(
                        torch.FloatTensor(self.demos_in_progress["env{}".format(i + 1)]["Demo"][tensor][demo_step]))

                # Insert demo logprob to self.step. Demo action prob is 1.0, so logprob is 0.0
                self.data[prl.LOGP][self.step][i].copy_(torch.zeros(1))

                # Set inserted data mask to 0.0 only if Intrinsic demo
                if prl.MASK in self.data.keys():
                    if self.demos_in_progress["env{}".format(i + 1)]["DemoType"] == "Intrinsic":
                        self.data[prl.MASK][self.step][i].copy_(torch.zeros(1))

                # Insert other tensors predicted in the forward pass
                for tensor in (prl.IREW, prl.VAL, prl.IVAL):

                    if tensor in algo_data.keys():
                        self.data[tensor][self.step][i].copy_(algo_data[tensor][num])

                    if tensor == prl.IREW and prl.IREW in algo_data.keys():
                        # Track the cumulative sum of intrinsic rewards of the demo
                        self.demos_in_progress["env{}".format(i + 1)]["CumIntrinsicReward"] += \
                            algo_data[prl.IREW][num].item()
                        # Track the max intrinsic reward of the demo
                        self.demos_in_progress["env{}".format(i + 1)]["MaxIntrinsicReward"] = max(
                            self.demos_in_progress["env{}".format(i + 1)]["MaxIntrinsicReward"],
                            algo_data[prl.IREW][num].item())

                # Update demo_in_progress variables
                self.demos_in_progress["env{}".format(i + 1)]["Step"] += 1
                self.demos_in_progress["env{}".format(i + 1)][prl.RHS] = {
                    k: rhs2[k][num].reshape(1, -1) for k in rhs2.keys()}
                self.inserted_samples += 1

                # Handle end of agent_demos
                if demo_step == self.demos_in_progress["env{}".format(i + 1)]["DemoLength"] - 1:

                    # If intrinsic demo
                    if self.demos_in_progress["env{}".format(i + 1)]["DemoType"] == "Intrinsic":
                        for intrinsic_demo in self.intrinsic_demos:
                            # If demo still in buffer, update IntrinsicReward
                            if self.demos_in_progress["env{}".format(i + 1)]["Demo"]["ID"] == intrinsic_demo["ID"]:
                                intrinsic_demo["IntrinsicReward"] = \
                                    self.eta * self.demos_in_progress["env{}".format(i + 1)]["MaxIntrinsicReward"] + \
                                    (1 - self.eta) * (self.demos_in_progress["env{}".format(i + 1)]["CumIntrinsicReward"] /
                                                      self.demos_in_progress["env{}".format(i + 1)]["DemoLength"])

                    # Randomly sample new agent_demos if last agent_demos has finished
                    self.sample_demo(env_id=i)

                else:

                    # Insert demo done2 tensor to self.step + 1
                    self.data[prl.DONE][self.step + 1][i].copy_(torch.zeros(1))

                    # Insert demo obs2 tensor to self.step + 1
                    obs2 = torch.roll(all_obs[num:num + 1], -self.num_channels_obs, dims=1).squeeze(0)
                    obs2[-self.num_channels_obs:].copy_(torch.FloatTensor(
                        self.demos_in_progress["env{}".format(i + 1)]["Demo"][prl.OBS][demo_step + 1]))
                    self.data[prl.OBS][self.step + 1][i].copy_(obs2)

                    # Insert demo rhs2 tensor to self.step + 1
                    if self.recurrent_actor:
                        for k in self.data[prl.RHS]:
                            self.data[prl.RHS][k][self.step + 1][i].copy_(rhs2[k][num].squeeze())

        self.step = (self.step + 1) % self.max_size
        self.size = min(self.size + 1, self.max_size)

    def track_potential_demos(self, sample):
        """ Tracks current episodes looking for potential agent_demos """

        for i in range(self.num_envs):

            # Copy transition
            for tensor in self.demos_data_fields:
                if tensor in (prl.OBS, ):
                    self.potential_demos["env{}".format(i + 1)][tensor].append(copy.deepcopy(
                        sample[tensor][i, -self.num_channels_obs:]).cpu().numpy().astype(self.demo_dtypes[tensor]))
                else:
                    self.potential_demos["env{}".format(i + 1)][tensor].append(
                        copy.deepcopy(sample[tensor][i]).cpu().numpy().astype(self.demo_dtypes[tensor]))

            # Update cumulative intrinsic reward
            if prl.IREW in sample.keys():

                # Track the cumulative sum of intrinsic rewards of the demo
                self.potential_demos_cumsum_int["env{}".format(i + 1)] += sample[prl.IREW][i].cpu().item()

                # Track the max intrinsic reward of the demo
                self.potential_demos_max_int["env{}".format(i + 1)] = max(
                    self.potential_demos_max_int["env{}".format(i + 1)], sample[prl.IREW][i].cpu().item())

            # Handle end of episode
            if sample[prl.DONE2][i] == 1.0:

                # Get candidate agent_demos
                potential_demo = {}
                for tensor in self.demos_data_fields:
                    potential_demo[tensor] = np.stack(self.potential_demos["env{}".format(i + 1)][tensor])
                    if tensor == prl.REW:
                        nonzero = np.flatnonzero(potential_demo[tensor] > 0.0)
                        last_reward = len(potential_demo[tensor]) if len(nonzero) == 0 else np.max(nonzero)

                # Compute accumulated reward
                episode_reward = potential_demo[prl.REW].sum()
                potential_demo["ID"] = str(uuid.uuid4())
                potential_demo["TotalReward"] = episode_reward
                potential_demo["DemoLength"] = potential_demo[prl.ACT].shape[0]

                # Consider candidate agent_demos for agent_demos reward
                if self.max_reward_demos > 0:

                    if episode_reward >= self.reward_threshold:

                        # Cut off data after last reward
                        for tensor in self.demos_data_fields:
                            potential_demo[tensor] = potential_demo[tensor][0:last_reward + 1]
                        potential_demo["DemoLength"] = potential_demo[prl.ACT].shape[0]

                        # Add agent_demos to reward buffer
                        self.reward_demos.append(potential_demo)

                        # Check if buffers are full
                        self.check_demo_buffer_capacity()

                        # Update reward_threshold.
                        self.reward_threshold = min([d["TotalReward"] for d in self.reward_demos])

                        # Update max demo reward
                        self.max_demo_reward = max([d["TotalReward"] for d in self.reward_demos])

                if self.max_intrinsic_demos > 0:

                    # Set potential demo cumulative intrinsic reward
                    episode_ireward = self.eta * self.potential_demos_max_int["env{}".format(i + 1)] + (1 - self.eta) * (
                            self.potential_demos_cumsum_int["env{}".format(i + 1)] / potential_demo["DemoLength"])
                    potential_demo["IntrinsicReward"] = episode_ireward

                    if episode_ireward >= self.intrinsic_threshold or len(self.intrinsic_demos) < self.max_intrinsic_demos:

                        # Add agent_demos to intrinsic buffer
                        self.intrinsic_demos.append(potential_demo)

                        # Check if buffer is full
                        self.check_demo_buffer_capacity()

                        # Update intrinsic_threshold
                        self.intrinsic_threshold = min([p["IntrinsicReward"] for p in self.intrinsic_demos])

                # Reset potential agent_demos dict
                for tensor in self.demos_data_fields:
                    self.potential_demos["env{}".format(i + 1)][tensor] = []
                    self.potential_demos_max_int["env{}".format(i + 1)] = 0.0
                    self.potential_demos_cumsum_int["env{}".format(i + 1)] = 0.0

    def load_demo(self, demo_path):
        """Loads and returns a environment demonstration."""

        # Load demos tensors
        demo = np.load(demo_path)
        new_demo = {k: {} for k in self.demos_data_fields}

        if int(demo["FrameSkip"]) != self.frame_skip:
            raise ValueError(
                "Env and demo with different frame skip!")

        # Add action, obs, rew
        new_demo[prl.ACT] = demo[prl.ACT]
        new_demo[prl.OBS] = demo[prl.OBS]
        new_demo[prl.REW] = demo[prl.REW]

        new_demo.update({
            "ID": str(uuid.uuid4()),
            "DemoLength": demo[prl.ACT].shape[0],
            "TotalReward": new_demo[prl.REW].sum(),
            "IntrinsicReward": 1000,
        })

        return new_demo

    def load_initial_demos(self):
        """
        Load initial demonstrations.

        Warning: make sure the frame_skip and frame_stack hyper-parameters are
        the same as those used to record the demonstrations!
        """

        loaded_int_demos, loaded_reward_demos = [], []

        initial_reward_demos = glob.glob(self.initial_reward_demos_dir + '/*.npz') if self.initial_reward_demos_dir else []
        initial_int_demos = glob.glob(self.initial_int_demos_dir + '/*.npz') if self.initial_int_demos_dir else []

        max_reward_demos = max(0, int(self.max_demos * (self.rho / (self.rho + self.phi))))
        max_int_demos = int(self.max_demos * (self.phi / (self.rho + self.phi)))

        for num, demo_file in enumerate(initial_reward_demos):

            if (num + 1) > max_reward_demos:
                continue

            try:

                new_demo = self.load_demo(demo_file)
                self.reward_demos.append(new_demo)
                loaded_reward_demos.append(new_demo["ID"])

            except Exception:
                print("Failed to load agent demo!")

        for num, demo_file in enumerate(initial_int_demos):

            if (num + 1) > max_int_demos:
                continue

            try:

                new_demo = self.load_demo(demo_file)
                self.intrinsic_demos.append(new_demo)
                loaded_int_demos.append(new_demo["ID"])

            except Exception:
                print("Failed to load intrinsic demo!")

        self.num_loaded_reward_demos, self.num_loaded_int_demos = len(loaded_reward_demos), len(loaded_int_demos)
        self.loaded_reward_demos, self.loaded_int_demos = tuple(loaded_reward_demos), tuple(loaded_int_demos)

        print("\nLOADED {} AGENT DEMOS AND {} INTRINSIC DEMOS".format(
            self.num_loaded_reward_demos, self.num_loaded_int_demos))

    def load_supplementary_demos(self):
        """
        Load demonstrations found in the self.supplementary_demos (if any).
        Warning: make sure the environment frame_skip and frame_stack hyperparameters are
        the same as those used in the demonstrations!
        """

        # Create supp dir if necessary
        if not os.path.exists(self.supplementary_demos_dir):
            os.makedirs(self.supplementary_demos_dir, exist_ok=True)

        num_loaded_supplementary_demos = 0
        supplementary_demos = glob.glob(self.supplementary_demos_dir + '/*.npz') if self.supplementary_demos_dir else []

        for demo_file in supplementary_demos:

            if demo_file not in self.supplementary_demos_loaded:

                self.supplementary_demos_loaded.append(demo_file)
                try:

                    new_demo = self.load_demo(demo_file)
                    self.reward_demos.append(new_demo)
                    self.intrinsic_demos.append(new_demo)
                    num_loaded_supplementary_demos += 1

                    # Check if buffer is full
                    self.check_demo_buffer_capacity()

                    # Update reward_threshold.
                    self.reward_threshold = min([d["TotalReward"] for d in self.reward_demos])

                    # Update intrinsic_threshold
                    self.intrinsic_threshold = min([p["IntrinsicReward"] for p in self.intrinsic_demos])

                    # Update max demo reward
                    self.max_demo_reward = max([d["TotalReward"] for d in self.reward_demos])

                except Exception:
                    print("Failed to load supplementary demo!")

        # Check if buffer is full
        self.check_demo_buffer_capacity()

        if num_loaded_supplementary_demos > 0:
            print("LOADED {} SUPPLEMENTARY DEMOS\n".format(num_loaded_supplementary_demos))

    def sample_demo(self, env_id):
        """With probability rho insert reward reward_demos, with probability phi insert intrinsic reward_demos."""

        # Reset reward_demos tracking variables
        self.demos_in_progress["env{}".format(env_id + 1)]["Step"] = 0
        self.demos_in_progress["env{}".format(env_id + 1)][prl.RHS] = None

        # Sample episode type
        episode_source = np.random.choice(["reward_demo", "intrinsic_demo", "env"],
            p=[self.rho, self.phi, 1.0 - self.rho - self.phi])

        if episode_source == "reward_demo" and len(self.reward_demos) > 0:

            # Randomly select a reward demo
            selected = np.random.choice(range(len(self.reward_demos)))
            demo = copy.deepcopy(self.reward_demos[selected])
            demo_type = "Reward"

        elif episode_source == "intrinsic_demo" and len(self.intrinsic_demos) > 0:

            # randomly select a intrinsic demo
            probs = np.array([(p["IntrinsicReward"]) ** self.alpha for p in self.intrinsic_demos])
            probs += 1e-10
            probs = (probs / probs.sum()).astype(np.float32)
            selected = np.random.choice(range(len(self.intrinsic_demos)), p=probs)
            demo = copy.deepcopy(self.intrinsic_demos[selected])
            demo_type = "Intrinsic"

        else:
            demo = None
            demo_type = None

        # Define demo and demo_type
        self.demos_in_progress["env{}".format(env_id + 1)]["Demo"] = demo
        self.demos_in_progress["env{}".format(env_id + 1)]["DemoType"] = demo_type

        # Set done to True
        self.data[prl.DONE][self.step + 1][env_id].copy_(torch.ones(1).to(self.device))

        # Set initial rhs to zeros
        if self.recurrent_actor:
            for k in self.data[prl.RHS]:
                self.data[prl.RHS][k][self.step + 1][env_id].fill_(0.0)

        if demo:

            # Set reward_demos length
            self.demos_in_progress["env{}".format(env_id + 1)]["DemoLength"] = demo["DemoLength"]

            # Set reward_demos MaxIntrinsicReward and CumIntrinsicReward to 0.0
            self.demos_in_progress["env{}".format(env_id + 1)]["CumIntrinsicReward"] = 0.0
            self.demos_in_progress["env{}".format(env_id + 1)]["MaxIntrinsicReward"] = 0.0

            # Set next buffer obs to be the starting demo obs
            for k in range(self.frame_stack):
                self.data[prl.OBS][self.step + 1][env_id][
                k * self.num_channels_obs:(k + 1) * self.num_channels_obs].copy_(torch.FloatTensor(
                    self.demos_in_progress["env{}".format(env_id + 1)]["Demo"][prl.OBS][0]))

        else:
            # Reset `i-th` environment as set next buffer obs to be the starting episode obs
            self.data[prl.OBS][self.step + 1][env_id].copy_(self.envs.reset_single_env(env_id=env_id).squeeze())

            # Reset potential reward_demos dict
            for tensor in self.demos_data_fields:
                self.potential_demos["env{}".format(env_id + 1)][tensor] = []
                self.potential_demos_max_int["env{}".format(env_id + 1)] = 0.0
                self.potential_demos_cumsum_int["env{}".format(env_id + 1)] = 0.0

    def check_demo_buffer_capacity(self):
        """
        Check total amount of reward_demos. If total amount of reward_demos exceeds
        self.max_demos, pop reward_demos.
        """

        # First pop intrinsic reward_demos with lowest CumulativeIntrinsicReward
        if len(self.intrinsic_demos) > self.max_intrinsic_demos:
            for _ in range(len(self.intrinsic_demos) - self.max_intrinsic_demos):
                del self.intrinsic_demos[np.array([p["IntrinsicReward"] for p in self.intrinsic_demos]).argmin()]

        # Second pop reward reward_demos with reward lower than self.max_demo_reward
        if len(self.reward_demos) > self.max_reward_demos:
            for _ in range(len(self.reward_demos) - self.max_reward_demos):
                rewards = np.array([p[prl.REW].sum() for p in self.reward_demos])
                if rewards[np.argmin(rewards)] < self.max_demo_reward:
                    del self.reward_demos[np.argmin(rewards)]

        # Finally eject longest reward_demos of those with reward self.max_demo_reward
        if len(self.reward_demos) > self.max_reward_demos:
            for _ in range(len(self.reward_demos) - self.max_reward_demos):
                lengths = np.array([p[prl.OBS].shape[0] for p in self.reward_demos])
                del self.reward_demos[np.argmax(lengths)]

    def save_demos(self):
        """
        Saves the top `num_rewards_demos` agent_demos from the reward agent_demos buffer and
        the top `num_intrinsic_demos` agent_demos from the intrinsic agent_demos buffer.
        """

        if self.target_reward_demos_dir:

            # Create target dir for reward agent_demos if necessary
            if not os.path.exists(self.target_reward_demos_dir):
                os.makedirs(self.target_reward_demos_dir, exist_ok=True)

            # Rank agent agent_demos according to episode reward
            reward_ranking = np.flip(np.array([d["TotalReward"] for d in self.reward_demos]).argsort())

            # Save agent reward agent_demos
            saved_demos = 0
            for demo_pos in reward_ranking:

                if saved_demos == self.num_reward_demos_to_save:
                    break

                filename = "reward_demo_{}".format(saved_demos + 1)

                save_data = {
                    prl.OBS: np.array(self.reward_demos[demo_pos][prl.OBS]).astype(self.demo_dtypes[prl.OBS]),
                    prl.REW: np.array(self.reward_demos[demo_pos][prl.REW]).astype(self.demo_dtypes[prl.REW]),
                    prl.ACT: np.array(self.reward_demos[demo_pos][prl.ACT]).astype(self.demo_dtypes[prl.ACT]),
                    "FrameSkip": self.frame_skip}

                if prl.EMBED in self.data.keys() and prl.EMBED in self.demos_data_fields:
                    save_data.update({prl.EMBED: np.array(self.reward_demos[demo_pos][prl.EMBED])})

                np.savez(os.path.join(self.target_reward_demos_dir, filename), **save_data),
                saved_demos += 1

        if self.target_int_demos_dir:

            # Create target dir for reward agent_demos if necessary
            if not os.path.exists(self.target_int_demos_dir):
                os.makedirs(self.target_int_demos_dir, exist_ok=True)

            # Rank agent agent_demos according to episode reward
            intrinsic_ranking = np.flip(np.array([d["IntrinsicReward"] for d in self.intrinsic_demos]).argsort())

            # Save agent intrinsic agent_demos
            saved_demos = 0
            for demo_pos in intrinsic_ranking:

                if saved_demos == self.num_int_demos_to_save:
                    break

                filename = "intrinsic_demo_{}".format(saved_demos + 1)

                save_data = {
                    prl.OBS: np.array(self.intrinsic_demos[demo_pos][prl.OBS]).astype(self.demo_dtypes[prl.OBS]),
                    prl.REW: np.array(self.intrinsic_demos[demo_pos][prl.REW]).astype(self.demo_dtypes[prl.REW]),
                    prl.ACT: np.array(self.intrinsic_demos[demo_pos][prl.ACT]).astype(self.demo_dtypes[prl.ACT]),
                    "FrameSkip": self.frame_skip}

                if prl.EMBED in self.data.keys() and prl.EMBED in self.demos_data_fields:
                    save_data.update({prl.EMBED: np.array(self.intrinsic_demos[demo_pos][prl.EMBED])})

                np.savez(os.path.join(self.target_int_demos_dir, filename), **save_data)
                saved_demos += 1



import gymnasium as gym
from minigrid.wrappers import RGBImgPartialObsWrapper, ImgObsWrapper, ReseedWrapper
from pytorchrl.envs.minigrid.custom_environments import register_custom_minigrid_envs
from pytorchrl.envs.minigrid.wrappers import RewardShapeWrapper, Gymnasium2GymWrapper

# Register custom environments
register_custom_minigrid_envs()


def minigrid_train_env_factory(env_id, index_worker=0, index_env=0, seed=None):
    """
    Create train MiniGrid environment.

    Parameters
    ----------
    env_id : str
        Environment name.
    index_worker : int
        Index of the worker running this environment.
    index_env : int
        Index of this environment withing the vector of environments.

    Returns
    -------
    env : gym.Env
        Train environment.
    """

    env = gym.make(env_id, tile_size=32)

    env = RewardShapeWrapper(env)

    if seed:
        # Fix environment seed
        env = ReseedWrapper(env, seeds=[seed])

    # Get pixel observations
    env = RGBImgPartialObsWrapper(env, tile_size=8)

    # Get rid of the 'mission' field
    env = ImgObsWrapper(env)

    # Translate to gym interface
    env = Gymnasium2GymWrapper(env)

    return env


def minigrid_test_env_factory(env_id, index_worker=0, index_env=0, seed=None):
    """
    Create test MiniGrid environment.

    Parameters
    ----------
    env_id : str
        Environment name.
    index_worker : int
        Index of the worker running this environment.
    index_env : int
        Index of this environment withing the vector of environments.

    Returns
    -------
    env : gym.Env
        Test environment.
    """

    env = gym.make(env_id, tile_size=32)

    env = RewardShapeWrapper(env)

    if seed:
        # Fix environment seed
        env = ReseedWrapper(env, seeds=[seed])

    # Get pixel observations
    env = RGBImgPartialObsWrapper(env, tile_size=8)

    # Get rid of the 'mission' field
    env = ImgObsWrapper(env)

    # Translate to gym interface
    env = Gymnasium2GymWrapper(env)

    return env



from abc import ABC, abstractmethod


class Storage(ABC):
    """
    Base class for all storage components. It should serve as a template to
    create new Storage classes with new or extended features.
    """

    def __init__(self, size, device, actor, algorithm, *args):
        """
        Initialize Storage class.

        Parameters
        ----------
        size : int
            Storage capacity along time axis.
        device: torch.device
            CPU or specific GPU where data tensors will be placed and class
            computations will take place. Should be the same device where the
            actor model is located.
        actor : Actor
            Actor class instance.
        algorithm : Algorithm
            Algorithm class instance
        """
        raise NotImplementedError


    @classmethod
    @abstractmethod
    def create_factory(cls, size, *args):
        """
        Returns a function to create new Storage instances.

        Parameters
        ----------
        size : int
            Storage capacity along time axis.
        """
        raise NotImplementedError

    @abstractmethod
    def init_tensors(self, sample, *args):
        """
        Lazy initialization of data tensors from a sample.

        Parameters
        ----------
        sample : dict
            Data sample (containing all tensors of an environment transition)
        """
        raise NotImplementedError

    @abstractmethod
    def get_all_buffer_data(self, data_to_cpu=False, *args):
        """
        Return all currently stored data. If data_to_cpu, moves
        data tensors to cpu memory.
        """
        raise NotImplementedError

    @abstractmethod
    def insert_data_slice(self, new_data, *args):
        """
        Add new_data to the buffer stored data.

        Parameters
        ----------
        new_data : dict
            Dictionary of env transition samples to replace self.data with.
        """
        raise NotImplementedError

    @abstractmethod
    def reset(self, *args):
        """Set class counters to zero and remove stored data"""
        raise NotImplementedError

    @abstractmethod
    def insert_transition(self, sample, *args):
        """
        Store new transition sample.

        Parameters
        ----------
        sample : dict
            Data sample (containing all tensors of an environment transition)
        """
        raise NotImplementedError

    @abstractmethod
    def before_gradients(self, actor, algo, *args):
        """
        Steps required before updating actor policy model.

        Parameters
        ----------
        actor : Actor class
            An actor class instance.
        algo : Algo class
            An algorithm class instance.
        """
        raise NotImplementedError

    @abstractmethod
    def after_gradients(self, actor, algo, info, *args):
        """
        Steps required after updating actor policy model

        Parameters
        ----------
        actor : Actor class
            An actor class instance.
        algo : Algo class
            An algorithm class instance.
        info : dict
            Additional relevant info from gradient computation.

        Returns
        -------
        info : dict
            info dict updated with relevant info from Storage.
        """
        raise NotImplementedError

    @abstractmethod
    def generate_batches(self, num_mini_batch, mini_batch_size, num_epochs=1, *args):
        """
        Returns a batch iterator to update actor critic.

        Parameters
        ----------
        num_mini_batch : int
           Number mini batches per epoch.
        mini_batch_size : int
            Number of samples contained in each mini batch.
        num_epochs : int
            Number of epochs.
        shuffle : bool
            Whether to shuffle collected data or generate sequential

        Yields
        ------
        batch : dict
            Generated data batches.
        """
        raise NotImplementedError

    @abstractmethod
    def update_storage_parameter(self, parameter_name, new_parameter_value, *args):
        """
        If `parameter_name` is an attribute of the algorithm, change its value
        to `new_parameter_value value`.

        Parameters
        ----------
        parameter_name : str
            Attribute name
        new_parameter_value : int or float
            New value for `parameter_name`.
        """
        raise NotImplementedError
